#!/usr/bin/env python
"""@package RunLAIReadMultiple

@brief Command line driver program to run lairead utility to initializes 
vegetation carbon stores. Will, for each worldfile in the project: (1) run 
lairead to produce a redefine worldfile; (2) run RHESSys simulation for 
3-days to generate base worldfile.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2015, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import sys
import os
import argparse

from rhessysworkflows.command.exceptions import *
from rhessysworkflows.command.modelrun import LAIReadMultiple

if __name__ == "__main__":
    # Handle command line options
    parser = argparse.ArgumentParser(description='Run lairead utility to initializes vegetation carbon stores. ' +
                                                 'Will, for each worldfile in the project: (1) run lairead to ' +
                                                 'produce a redefine worldfile; (2) run RHESSys simulation for 3-days to generate base worldfile')
    parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                        help='The configuration file. Must define section "GRASS" and option "GISBASE"')
    parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                        help='The directory to which metadata, intermediate, and final files should be saved')
    parser.add_argument('--topmodel', dest='topmodel', required=False, action='store_true',
                        help='Run RHESSys in topmodel mode when running lairead')
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help='Print detailed information about what the program is doing')
    args = parser.parse_args()
    
    configFile = None
    if args.configfile:
        configFile = args.configfile
        
    command = LAIReadMultiple(args.projectDir, configFile)
    
    exitCode = os.EX_OK
    try: 
        command.run(verbose=args.verbose, topmodel=args.topmodel)
    except CommandException as e:
        print(str(e))
        exitCode = os.EX_DATAERR
    
    sys.exit(exitCode)


#!/usr/bin/env python
"""@package RegisterLandcoverReclassRules

@brief Generate landcover raster maps reclassification rules for use with GenerateLandcoverMaps.py.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'SCRIPT', 'ETC'
   
2. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory:
   landcover_type [required only if --generateKnownRules is supplied]
   
3. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   rhessys_dir

Post conditions
---------------
1. Will write the following entry(ies) to the RHESSys section of metadata associated with the project directory:
   landcover_road_rule
   landcover_impervious_rule
   landcover_landuse_rule
   landcover_stratum_rule
   landcover_lai_rule [if LAI flag is specified]

Usage:
@code
RegisterLandcoverReclassRules.py -p /path/to/project_dir
@endcode

@note EcoHydroWorkflowLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified.
"""
import os, sys, errno, shutil
import argparse
import textwrap

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import RHESSysPaths

# Handle command line options
parser = argparse.ArgumentParser(description='Generate landcover maps in GRASS GIS')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
group = parser.add_mutually_exclusive_group(required=True)
group.add_argument('-k', '--generateKnownRules', dest='generateKnownRules', required=False, action='store_true',
                    help="Generate rules for known landcover type; known types include: %s" % (str(RHESSysMetadata.KNOWN_LC_TYPES), ) )
group.add_argument('-b', '--buildPrototypeRules', dest='buildPrototypeRules', required=False, action='store_true',
                    help='Write prototype landcover reclass rules to the project directory. You must edit these rules to match the classes in your own landcover data')
group.add_argument('-r', '--ruleDir', dest='ruleDir', required=False,
                    help="The directory where landcover reclass rules can be found; should contain these files %s" % (str(RHESSysMetadata.LC_RULES),) )
parser.add_argument('-l', '--includeLaiRules', dest='includeLaiRules', required=False, action='store_true',
                    help='Make LAI map')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

# Check for necessary information in metadata
metadata = RHESSysMetadata.readRHESSysEntries(context)
studyArea = RHESSysMetadata.readStudyAreaEntries(context)
manifest = RHESSysMetadata.readManifestEntries(context)

paths = RHESSysPaths(args.projectDir, metadata['rhessys_dir'])

# Get path of place to store reclass rules
projectDirRuleDir = paths.getReclassRulesDirectory()

# Write prototype landcover reclass rules to the project directory
if args.buildPrototypeRules:
    sys.stdout.write('Generating prototype landcover reclass rules...')
    sys.stdout.flush()
    # Road rule
    roadRulePath = os.path.join(projectDirRuleDir, RHESSysMetadata.LC_RULE_ROAD)
    with open(roadRulePath, 'w') as f:
        f.write('22 23 24 31 = 1 road\n')
        f.write('* = NULL\n')
    # Impervious rule
    imperviousRulePath = os.path.join(projectDirRuleDir, RHESSysMetadata.LC_RULE_IMPERVIOUS)
    with open(imperviousRulePath, 'w') as f:
        f.write('22 23 24 31 = 1 impervious\n')
        f.write('* = 0 pervious\n')
    # Landuse rule
    landuseRulePath = os.path.join(projectDirRuleDir, RHESSysMetadata.LC_RULE_LANDUSE)
    with open(landuseRulePath, 'w') as f:
        f.write('11 12 31 41 42 43 51 52 71 thru 74 90 95 = 1 undeveloped\n')
        f.write('81 82 = 2 agriculture\n')
        f.write('21 22 23 24 = 3 urban\n')
    # Stratum rule
    stratumRulePath = os.path.join(projectDirRuleDir, RHESSysMetadata.LC_RULE_STRATUM)
    with open(stratumRulePath, 'w') as f:
        f.write('11 12 21 22 23 24 31 73 74 82 = 1 nonveg\n')
        f.write('71 72 81 95 = 2 grass\n')
        f.write('41 43 90 = 3 deciduous\n')
        f.write('42 51 52 = 4 evergreen\n')
    if args.includeLaiRules:
        # LAI rule
        laiRulePath = os.path.join(projectDirRuleDir, RHESSysMetadata.LC_RULE_LAI)
        with open(laiRulePath, 'w') as f:
            f.write('11:12:0.0\n')
            f.write('21:24:0.0\n') 
            f.write('31:31:0.0\n') 
            f.write('73:74:0.0\n') 
            f.write('82:82:0.0\n')
            f.write('71:72:1.5\n')
            f.write('81:81:1.5\n')
            f.write('95:95:1.5\n')
            f.write('41:41:5.0\n')
            f.write('43:43:5.0\n')
            f.write('90:90:5.0\n')
            f.write('42:42:6.0\n')
            f.write('51:52:6.0\n')
            
    ruleDir = None
    sys.stdout.write('done\n')

# Generate rules for known landcover type
if args.generateKnownRules:
    # Check for necessary information in metadata
    if not 'landcover_type' in studyArea:
        sys.exit("Metadata in project directory %s does not contain a landcover type" % (context.projectDir,) )
    
    landcoverType = studyArea['landcover_type']
    if landcoverType not in RHESSysMetadata.KNOWN_LC_TYPES:
        sys.exit("Landcover type '%s' is not known, so rules cannot be generated." % (landcoverType,) )
    scriptEtc = context.config.get('SCRIPT', 'ETC')
    ruleDir = os.path.join(scriptEtc, landcoverType)

# Import rules from elsewhere
if args.ruleDir:
    if not os.access(args.ruleDir, os.R_OK):
        sys.exit("Unable to read rule directory %s" % (ruleDir,) )
    ruleDir = os.path.abspath(args.ruleDir)

if ruleDir:
    sys.stdout.write(textwrap.fill("Importing landcover reclass rules from %s ..." % (ruleDir,) ) )
    sys.stdout.flush()
    # Copy rules into project directory
    roadRulePath = os.path.join(ruleDir, RHESSysMetadata.LC_RULE_ROAD)
    shutil.copy(roadRulePath, projectDirRuleDir)
    imperviousRulePath = os.path.join(ruleDir, RHESSysMetadata.LC_RULE_IMPERVIOUS)
    shutil.copy(imperviousRulePath, projectDirRuleDir)
    landuseRulePath = os.path.join(ruleDir, RHESSysMetadata.LC_RULE_LANDUSE)
    shutil.copy(landuseRulePath, projectDirRuleDir)
    stratumRulePath = os.path.join(ruleDir, RHESSysMetadata.LC_RULE_STRATUM)
    shutil.copy(stratumRulePath, projectDirRuleDir)
    if args.includeLaiRules:
        laiRulePath = os.path.join(ruleDir, RHESSysMetadata.LC_RULE_LAI)
        shutil.copy(laiRulePath, projectDirRuleDir)
    sys.stdout.write('done\n')
    
# Write metadata
RHESSysMetadata.writeRHESSysEntry(context, 'landcover_stratum_rule', os.path.join(RHESSysMetadata.RULES_DIR, os.path.basename(stratumRulePath)) )
RHESSysMetadata.writeRHESSysEntry(context, 'landcover_landuse_rule', os.path.join(RHESSysMetadata.RULES_DIR, os.path.basename(landuseRulePath)) )
RHESSysMetadata.writeRHESSysEntry(context, 'landcover_impervious_rule', os.path.join(RHESSysMetadata.RULES_DIR, os.path.basename(imperviousRulePath)) )
RHESSysMetadata.writeRHESSysEntry(context, 'landcover_road_rule', os.path.join(RHESSysMetadata.RULES_DIR, os.path.basename(roadRulePath)) )
if args.includeLaiRules:
    RHESSysMetadata.writeRHESSysEntry(context, 'landcover_lai_rule', os.path.join(RHESSysMetadata.RULES_DIR, os.path.basename(laiRulePath)) )

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)


#!/usr/bin/env python
"""@package RunLAIRead

@brief Run lairead utility to initializes vegetation carbon stores. Will: (1) run lairead to
produce a redefine worldfile; (2) run RHESSys simulation for 3-days to generate base worldfile

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   grass_dbase
   grass_location
   grass_mapset
   rhessys_dir
   rhessys_bin
   lairead_bin
   worldfile_zero
   surface_flowtable
   subsurface_flowtable
   allometric_table

2. The following metadata entry(ies) must be present in the GRASS section of the metadata associated with the project directory:
   basin_rast
   dem_rast
   hillslope_rast
   zone_rast
   patch_rast
   stratum_rast
   lai_rast
   
Post conditions
---------------
1. Worldfile with initialized vegetation carbon stores will be saved to the appropriate location of the project directory

2. Will write the following entry(ies) to the RHESSys section of metadata associated with the project directory:
   worldfile
   tec_lairead

Usage:
@code
RunLAIRead.py -p /path/to/project_dir
@endcode

@note EcoHydroWorkflowLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import argparse
import datetime
from subprocess import *
import shutil

from ecohydrolib.grasslib import *

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import RHESSysPaths
from rhessysworkflows.rhessys import datetimeToString
from rhessysworkflows.rhessys import generateCommandString
from rhessysworkflows.worldfileio import getClimateBaseStationFilenames
from rhessysworkflows.climateio import getStartAndEndDateForClimateStation

# Handle command line options
parser = argparse.ArgumentParser(description='Run lairead utility to initializes vegetation carbon stores. Will: (1) run lairead to ' +
                                             'produce a redefine worldfile; (2) run RHESSys simulation for 3-days to generate base worldfile')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                    help='Print detailed information about what the program is doing')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

# Check for necessary information in metadata
grassMetadata = RHESSysMetadata.readGRASSEntries(context)
if not 'basin_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a basin raster" % (context.projectDir,))
if not 'dem_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a DEM raster" % (context.projectDir,))
if not 'hillslope_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a hillslope raster" % (context.projectDir,))  
if not 'zone_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a zone raster" % (context.projectDir,)) 
if not 'patch_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a patch raster" % (context.projectDir,)) 
if not 'stratum_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with a stratum raster" % (context.projectDir,)) 
if not 'lai_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS dataset with an LAI raster" % (context.projectDir,)) 

metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))
if not 'rhessys_dir' in metadata:
    sys.exit("Metadata in project directory %s does not contain a RHESSys directory" % (context.projectDir,))
if not 'rhessys_bin' in metadata:
    sys.exit("Metadata in project directory %s does not contain a RHESSys binary" % (context.projectDir,))
if not 'lairead_bin' in metadata:
    sys.exit("Metadata in project directory %s does not contain a lairead executable" % (context.projectDir,))
if not 'worldfile_zero' in metadata:
    sys.exit("Metadata in project directory %s does not contain an initial worldfile" % (context.projectDir,))
if not 'surface_flowtable' in metadata:
    sys.exit("Metadata in project directory %s does not contain a surface flowtable" % (context.projectDir,))
if not 'subsurface_flowtable' in metadata:
    sys.exit("Metadata in project directory %s does not contain a subsurface flowtable" % (context.projectDir,))
if not 'allometric_table' in metadata:
    sys.exit("Metadata in project directory %s does not contain an allometric table" % (context.projectDir,))

rhessysDir = metadata['rhessys_dir']
paths = RHESSysPaths(args.projectDir, rhessysDir)

worldfileZero = metadata['worldfile_zero']
headerZero = "%s.hdr" % (worldfileZero,)
worldfileZeroPath = os.path.join(context.projectDir, worldfileZero)
worldfileDir = os.path.dirname(worldfileZeroPath)

# Set up GRASS environment
modulePath = context.config.get('GRASS', 'MODULE_PATH')
grassDbase = os.path.join(context.projectDir, metadata['grass_dbase'])
grassConfig = GRASSConfig(context, grassDbase, metadata['grass_location'], metadata['grass_mapset'])
grassLib = GRASSLib(grassConfig=grassConfig)

## Make sure mask and region are properly set
demRast = grassMetadata['dem_rast']
result = grassLib.script.run_command('g.region', rast=demRast)
if result != 0:
    sys.exit("g.region failed to set region to DEM, returning %s" % (result,))

basinRast = grassMetadata['basin_rast']
result = grassLib.script.run_command('r.mask', flags='o', input=basinRast, maskcats='1')
if result != 0:
    sys.exit("r.mask failed to set mask to basin, returning %s" % (result,))

## 1. Determine legal simulation start and date from climate data 
# Read first climate station from worldfile
headerZeroPath = os.path.join(context.projectDir, headerZero)
stations = getClimateBaseStationFilenames(headerZeroPath)
assert( len(stations) )
firstStationPath = os.path.normpath( os.path.join(paths.RHESSYS_DIR, stations[0]) )
if args.verbose:
    sys.stdout.write("First climate station in worldfile: %s\n" % (firstStationPath,) )

# Read climate timeseries for start and end date, write to metadata
(startDate, endDate) = getStartAndEndDateForClimateStation(firstStationPath, paths)
if args.verbose:
    sys.stdout.write("start date: %s, end date: %s\n" % ( str(startDate), str(endDate) ) )
fourDays = datetime.timedelta(days=4)
if endDate - startDate < fourDays:
    sys.exit("Climate time-series defined by station %s is too short to run lairead (less than four-days long)" %
             (firstStationPath,) )

## 2. Run LAI read to generate redefine worldfile
tecDurRedef = datetime.timedelta(days=1)
tecRedef = startDate + tecDurRedef
laireadPath = os.path.join(context.projectDir, metadata['lairead_bin'])
oldWorldPath = os.path.join(context.projectDir, worldfileZero)
redefWorldName = "%s.Y%dM%dD%dH%d" % \
    (worldfileZero, tecRedef.year, tecRedef.month, tecRedef.day, tecRedef.hour)
redefWorldPath = os.path.join(context.projectDir, redefWorldName)
allomPath = os.path.join(context.projectDir, metadata['allometric_table'])

p = grassLib.script.pipe_command(laireadPath, old=oldWorldPath, redef=redefWorldPath,
                                      allom=allomPath, lai=grassMetadata['lai_rast'],
                                      vegid=grassMetadata['stratum_rast'],
                                      zone=grassMetadata['zone_rast'],
                                      hill=grassMetadata['hillslope_rast'],
                                      patch=grassMetadata['patch_rast'],
                                      mask=grassMetadata['basin_rast'])
(stdoutStr, stderrStr) = p.communicate() 
result = p.returncode
if result != 0:
    sys.stdout.write(stdoutStr)
    sys.exit("\nlairead failed, returning %s" % (result,))

## 3. Write TEC file for redefining the initial flow table
##    Redefine on the second day of the simulation, write output
##    on the third day
tecName = 'tec.lairead'
tecPath = os.path.join(paths.RHESSYS_TEC, tecName)
tecDurOutput = datetime.timedelta(days=2)
tecOutput = startDate + tecDurOutput

f = open(tecPath, 'w')
f.write("%s redefine_world%s" % 
        (datetimeToString(tecRedef), os.linesep) )
f.write("%s output_current_state%s" %
        (datetimeToString(tecOutput), os.linesep) )
f.close()
RHESSysMetadata.writeRHESSysEntry( context, 'tec_lairead', paths.relpath(tecPath) )

## 4. Run RHESSys for the first 4 legal days with redefine TEC
rhessysStart = startDate
rhessysDur = datetime.timedelta(days=3)
rhessysEnd = startDate + rhessysDur
surfaceFlowtablePath = os.path.join(context.projectDir, metadata['surface_flowtable'])
subSurfaceFlowtablePath = os.path.join(context.projectDir, metadata['subsurface_flowtable'])
rhessysBinPath = os.path.join(context.projectDir, metadata['rhessys_bin'])

rhessysCmd = generateCommandString(rhessysBinPath, None,
                                   rhessysStart, rhessysEnd,
                                   tecPath, oldWorldPath,
                                   subSurfaceFlowtablePath, surfaceFlowtablePath)
if args.verbose:
    print(rhessysCmd)
sys.stdout.write('\nRunning RHESSys to redefine worldfile with vegetation carbon stores...')
sys.stdout.flush()

cmdArgs = rhessysCmd.split()
process = Popen(cmdArgs, cwd=paths.RHESSYS_DIR, stdout=PIPE, stderr=PIPE)
(process_stdout, process_stderr) = process.communicate()
if args.verbose:
    sys.stdout.write(process_stdout)
    sys.stderr.write(process_stderr)
if process.returncode != 0:
    sys.exit("\n\nRHESSys failed, returning %s" % (process.returncode,) )

sys.stdout.write('done\n')

## 5. Rename redefine worldfile, write to metadata
outputWorldName = "%s.Y%dM%dD%dH%d.state" % \
    (worldfileZero, tecOutput.year, tecOutput.month, tecOutput.day, tecOutput.hour)
outputWorldPath = os.path.join(context.projectDir, outputWorldName)

if not os.access(outputWorldPath, os.W_OK):
    sys.exit("Unable to find redefined worldfile %s" % (outputWorldPath,) )
    
newWorldName = 'world'
newWorldPath = os.path.join(paths.RHESSYS_WORLD, newWorldName)

shutil.move(outputWorldPath, newWorldPath)
if not os.path.exists(newWorldPath):
    sys.exit("Failed to copy redefined worldfile %s to %s" % (outputWorldPath, newWorldPath) )
RHESSysMetadata.writeRHESSysEntry( context, 'worldfile', paths.relpath(newWorldPath) )

# Copy world file header from init worldfile to final world file
header = "%s.hdr" % (newWorldName,)
headerPath = os.path.join(paths.RHESSYS_WORLD, header)
shutil.copyfile(headerZeroPath, headerPath)

sys.stdout.write('\n\nSuccessfully used lairead to initialize vegetation carbon stores.\n')

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)


#!/usr/bin/env python
"""@package ImportRasterMapIntoGRASS

@brief Import raster map already registered in metadata via EcohydroLib.RegisterRaster, into GRASS.
Raster type must be one of RHESSysMetadata.RASTER_TYPES.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GRASS', 'GISBASE'

2. One or more of the following metadata entry(ies) must be present in the manifest section of the metadata associated with the project directory:
   rhessysworkflows.metadata.RASTER_TYPES, currently:
   landcover
   roof_connectivity
   soil
   lai
   patch
   zone
   isohyet
   leafc
   rootdepth
   
3. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   grass_dbase
   grass_location
   grass_mapset
   
Post conditions
---------------
1. Will write one or more the following entry(ies) to the GRASS section of metadata associated with the project directory:
   dem_rast
   landcover_rast
   roof_connectivity_rast
   soil_rast
   lai_rast
   patch_rast
   zone_rast

2. Will delete one or more of the following entry(ies) from the RHESSys section of metadata associated with the project directory:
   stratum_defs
   landuse_defs
   soil_defs

Usage:
@code
ImportRasterMapIntoGRASS.py -t RASTER_TYPE_1 [RASTER_TYPE_2 ... RASTER_TYPE_N] -p /path/to/project_dir
@endcode

@note EcoHydroWorkflowLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import os, sys
import argparse
import textwrap

from ecohydrolib.grasslib import *

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import RHESSysPaths

RESAMPLE_METHODS = list(GRASS_RASTER_RESAMPLE_METHODS)
RESAMPLE_METHODS.insert(0, 'none')

# Handle command line options
typeChoices = list(RHESSysMetadata.RASTER_TYPES)
typeChoices.insert(0, 'all')
parser = argparse.ArgumentParser(description='Import raster map already registered in metadata via EcohydroLib.RegisterRaster, into GRASS. Raster type must be one of RHESSysMetadata.RASTER_TYPES.')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-t', '--type', dest='types', required=True, nargs='+', choices=typeChoices,
                    help='The type of raster dataset to import, or all if all types should be imported.')
parser.add_argument('-m', '--method', dest='method', required=True, 
                    nargs='+', choices=RESAMPLE_METHODS,
                    help='The method to use to resample each raster.')
parser.add_argument('--integer', dest='integer', required=False, action='store_true',
                    help='Transform raster to integer on import. Can only be used when importing a single raster type.')
parser.add_argument('--multiplier', dest='multiplier', required=False, type=int, default=1000,
                    help='Multiplier to use when tranforming raster values to integer')                    
parser.add_argument('--overwrite', dest='overwrite', action='store_true', required=False,
                    help='Overwrite existing datasets in the GRASS mapset.  If not specified, program will halt if a dataset already exists.')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if len(args.types) != len(args.method):
    sys.exit(textwrap.fill("%d raster type(s) specified for import, but %d resample method(s) supplied. Please specify a resample method for each raster type" % \
             (len(args.types), len(args.method)) ) )

if 'all' in args.types:
    typeList = RHESSysMetadata.RASTER_TYPES
    methodList = [args.method[0] for t in typeList]
else:
    typeList = args.types
    methodList = args.method

if args.integer:
    if len(typeList) > 1:
        sys.exit(textwrap.fill("Integer transformation can only be applied when importing a single raster type"))

## Check for necessary information in metadata
manifest = RHESSysMetadata.readManifestEntries(context)
metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))

grassMetadata = RHESSysMetadata.readGRASSEntries(context)
if not 'dem_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a DEM raster in a GRASS mapset" % (context.projectDir,)) 
demRast = grassMetadata['dem_rast']

# Set up GRASS environment
modulePath = context.config.get('GRASS', 'MODULE_PATH')
grassDbase = os.path.join(context.projectDir, metadata['grass_dbase'])
grassConfig = GRASSConfig(context, grassDbase, metadata['grass_location'], metadata['grass_mapset'])
grassLib = GRASSLib(grassConfig=grassConfig)

# Make sure mask and region are properly set
result = grassLib.script.run_command('r.mask', flags='r')
if result != 0:
    sys.exit("r.mask filed, returning %s" % (result,) )
result = grassLib.script.run_command('g.region', rast=demRast)
if result != 0:
    sys.exit("g.region failed to set region to DEM, returning %s" % (result,))

# Import raster maps into GRASS
for index, t in enumerate(typeList):
    if t in manifest:
        method = methodList[index]
        sys.stdout.write("Importing %s raster..." % (t,) )
        sys.stdout.flush()
        rasterPath = os.path.join(context.projectDir, manifest[t])
        if args.integer:
            importName = "%s_tmp" % (t,)
            overwrite = True
        else:
            importName = t 
            overwrite = args.overwrite
        result = grassLib.script.run_command('r.in.gdal', input=rasterPath, output=importName, overwrite=overwrite)
        if result != 0:
            sys.exit("Failed to import raster %s into GRASS dataset %s/%s, result:\n%s" % \
                     (importName, grassDbase, metadata['grass_location'], result) )
        if method != 'none':
            # Resample the raster to ensure it is the same resolution as the geographic region defined by the DEM
            sys.stdout.write("Resampling %s raster using method %s...\n" % (t, method) )
            sys.stdout.flush()
            result = grassLib.script.run_command('r.resamp.interp', input=importName, output=importName, method=method, overwrite=True)
            if result != 0:
                sys.exit("Failed to resample imported raster %s of GRASS dataset %s/%s, result:\n%s" % \
                         (t, grassDbase, metadata['grass_location'], result) )
        if args.integer:
            # Convert raster to integer
            sys.stdout.write("Converting %s raster to integer using multiplier %s...\n" % (t, str(args.multiplier)) )
            sys.stdout.flush()
            mapcalcCmd = "%s=int(%s * %s)" % (t, importName, str(args.multiplier) )
            print(mapcalcCmd)
            result = grassLib.script.write_command('r.mapcalc', stdin=mapcalcCmd)
            if result != 0:
                sys.exit("Integer conversion failed for raster %s of GRASS dataset %s/%s, result:\n%s" % \
                         (t, grassDbase, metadata['grass_location'], result) )
            result = grassLib.script.run_command('g.remove', rast=importName)
            if result != 0:
                sys.exit("Failed to delete temporary raster %s of GRASS dataset %s/%s, result:\n%s" % \
                         (importName, grassDbase, metadata['grass_location'], result) )
        
        grassEntryKey = "%s_rast" % (t,)
        RHESSysMetadata.writeGRASSEntry(context, grassEntryKey, t)
        sys.stdout.write('done\n')
        # Invalidate metadata as necessary
        if t == RHESSysMetadata.RASTER_TYPE_LC:
            RHESSysMetadata.deleteRHESSysEntry(context, 'stratum_defs')
            RHESSysMetadata.deleteRHESSysEntry(context, 'landuse_defs')
        elif t == RHESSysMetadata.RASTER_TYPE_SOIL:
            RHESSysMetadata.deleteRHESSysEntry(context, 'soil_defs')

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)



#!/usr/bin/env python
"""@package CreateWorldfile

@brief Create RHESSys worldfile.  This will create an initial worldfile used for input into the
lairead utility, which initializes vegetation carbon stores.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   grass_dbase
   grass_location
   grass_mapset
   rhessys_dir
   g2w_bin
   rat_bin
   template

   
Post conditions
---------------
1. Worldfile will be created in the RHESSys folder of the project directory.

2. Will write the following entry(ies) to the RHESSys section of metadata associated with the project directory:
   worldfile_zero

Usage:
@code
CreateWorldfile.py -p /path/to/project_dir
@endcode

@note EcoHydroWorkflowLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import string
import re
import argparse
from subprocess import *

from ecohydrolib.grasslib import *
from ecohydrolib.spatialdata.utils import bboxFromString
from ecohydrolib.spatialdata.utils import calculateBoundingBoxCenter

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import RHESSysPaths
from rhessysworkflows.rhessys import readParameterFile

# Handle command line options
parser = argparse.ArgumentParser(description='Create RHESSys worldfile using GRASS GIS data and grass2world utility')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                    help='Print detailed information about what the program is doing')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

# Check for necessary information in metadata
grassMetadata = RHESSysMetadata.readGRASSEntries(context)
if not 'dem_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a DEM raster in a GRASS mapset" % (context.projectDir,)) 
if not 'soil_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a soil raster in a GRASS mapset" % (context.projectDir,))
if not 'patch_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a patch raster in a GRASS mapset" % (context.projectDir,))

metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))
if not 'rhessys_dir' in metadata:
    sys.exit("Metadata in project directory %s does not contain a RHESSys directory" % (context.projectDir,))
if not 'g2w_bin' in metadata:
    sys.exit("Metadata in project directory %s does not contain a grass2world executable" % (context.projectDir,))
if not 'rat_bin' in metadata:
    sys.exit("Metadata in project directory %s does not contain an AverageTables executable" % (context.projectDir,))
if not 'template' in metadata:
    sys.exit("Metadata in project directory %s does not contain an world template" % (context.projectDir,))

rhessysDir = metadata['rhessys_dir']
paths = RHESSysPaths(args.projectDir, rhessysDir)

# Set up GRASS environment
modulePath = context.config.get('GRASS', 'MODULE_PATH')
grassDbase = os.path.join(context.projectDir, metadata['grass_dbase'])
grassConfig = GRASSConfig(context, grassDbase, metadata['grass_location'], metadata['grass_mapset'])
grassLib = GRASSLib(grassConfig=grassConfig)

## Run grass2world
# Make sure mask and region are properly set
demRast = grassMetadata['dem_rast']
result = grassLib.script.run_command('g.region', rast=demRast)
if result != 0:
    sys.exit("g.region failed to set region to DEM, returning %s" % (result,))

basinRast = grassMetadata['basin_rast']
result = grassLib.script.run_command('r.mask', flags='o', input=basinRast, maskcats='1')
if result != 0:
    sys.exit("r.mask failed to set mask to basin, returning %s" % (result,))

templateFilename = os.path.basename( metadata['template'] )
templateFilepath = os.path.join( context.projectDir, metadata['template'] )

worldfileName = "%s_init" % (templateFilename.replace('template', 'world'), )
worldfilePath = os.path.join(paths.RHESSYS_WORLD, worldfileName)

g2wPath = os.path.join(context.projectDir, metadata['g2w_bin'])
g2wCommand = "%s -t %s -w %s" % (g2wPath, templateFilepath, worldfilePath)

# Make sure g2w can find rat
g2wEnv = dict(os.environ)
g2wEnv['PATH'] = paths.RHESSYS_BIN + os.pathsep + g2wEnv['PATH']

if args.verbose:
    print(g2wCommand)
sys.stdout.write("\nRunning grass2world from %s..." % (paths.RHESSYS_BIN,) )
sys.stdout.flush()
cmdArgs = g2wCommand.split()
process = Popen(cmdArgs, cwd=paths.RHESSYS_BIN, env=g2wEnv, 
                stdout=PIPE, stderr=PIPE)
(process_stdout, process_stderr) = process.communicate()
if args.verbose:
    sys.stdout.write(process_stdout)
    sys.stderr.write(process_stderr)
if process.returncode != 0:
    sys.exit("\n\ngrass2world failed, returning %s" % (process.returncode,) )    

RHESSysMetadata.writeRHESSysEntry(context, 'worldfile_zero', paths.relpath(worldfilePath) )

sys.stdout.write('\n\nFinished creating worldfile\n')

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)



#!/usr/bin/env python
# TODO: Add ability to plot all three plot types in one figure
# TODO: Add table plot type that print summary statistics for column
import sys
import argparse
import math

import numpy as np
import pandas as pd
import statsmodels.api as sm
import matplotlib.pyplot as plt
import matplotlib

from rhessysworkflows.rhessys import RHESSysOutput

PLOT_TYPE_STD = 'standard'
PLOT_TYPE_LOGY = 'logy'
PLOT_TYPE_CDF = 'cdf'
PLOT_TYPE_TABLE = 'table'
PLOT_TYPES = [PLOT_TYPE_STD, PLOT_TYPE_LOGY, PLOT_TYPE_CDF, PLOT_TYPE_TABLE]
PLOT_DEFAULT = PLOT_TYPE_STD


def plotTable(args, col_names, obs, data, ax):
    
    #import pdb; pdb.set_trace()
    
    startDate = data['streamflow'].index[0]
    endDate = data['streamflow'].index[-1]
    
    obsRunoffRat = np.sum( obs['observed'] ) / np.sum( data['precip'] )
    modRunoffRat = np.sum( data['streamflow']) / np.sum( data['precip'] )
    obsET = np.sum( data['precip'] ) - np.sum( obs['observed'] )
    summary = [ [ np.sum( data['precip'] ), np.sum( data['precip'] ) ],
                [ np.sum(obs['observed']), np.sum(data['streamflow']) ],
                [ obsET, np.sum( data['evap'] ) + np.sum( data['trans'] ) ],
                [ obsRunoffRat, modRunoffRat ]
              ]
    text = [ [ "%.2f" % num for num in summary[0] ],
             [ "%.2f" % num for num in summary[1] ],
             [ "%.2f" % num for num in summary[2] ],
             [ "%.2f" % num for num in summary[3] ]
           ]

    ax.axis('off')
    # Draw the table
    table = ax.table(cellText=text,
                     colWidths=[0.33, 0.33],
                     colLabels=col_names,
                     rowLabels=['Precip ($mm$)', 'Streamflow ($mm$)', 'ET ($mm$)', 'Runoff ratio'],
                     loc='center right')
    table.auto_set_font_size(False)
    table.set_fontsize(9)
    ax.text(0, 0.75, "Mass balance: %s - %s" % \
                 (startDate.strftime('%Y/%m/%d'), endDate.strftime('%Y/%m/%d') ),
            horizontalalignment='left',
            #verticalalignment='center',
            fontsize=10)
    

def plotGraph(args, plottype, obs, data, columns, min_x, max_x, ax, secondary=None,
              plotColor=True):
    
    if plotColor:
        obs_color = 'black'
        mod_color = 'green'
        second_color = 'blue'
    else:
        obs_color = 'black'
        mod_color = 'grey'
        second_color = 'black'
    
    if plottype == PLOT_TYPE_STD or \
       plottype == PLOT_TYPE_LOGY:
        x = obs.index
    elif plottype == PLOT_TYPE_CDF:
        x = np.linspace(min_x, max_x, num=1000 )
    
    # Plot observed values
    # Standard or log plot
    obs_y = obs
    if plottype == PLOT_TYPE_CDF:
        obs_ecdf = sm.distributions.ECDF(obs['observed'])
        obs_y = obs_ecdf(x)
    obs_plt = None
    if not args.supressObs:
        (obs_plt,) = ax.plot(x, obs_y, obs_color, linewidth=2)
        
    # Plot modeled values
    data_plt = []
    for c in columns:
        # Standard or log plot
        mod_y = data[c]
        if plottype == PLOT_TYPE_CDF:
            mod_ecdf = sm.distributions.ECDF(data[c])
            mod_y = mod_ecdf(x)
        (mod_plt,) = ax.plot(x, mod_y, color=mod_color, linewidth=1)
        data_plt.append(mod_plt)
    
    # X-axis
    if plottype == PLOT_TYPE_STD or \
       plottype == PLOT_TYPE_LOGY:
        num_years = len(x) / 365
        if num_years > 2:
            if num_years > 5:
                ax.xaxis.set_major_locator(matplotlib.dates.YearLocator())
            else:
                ax.xaxis.set_major_locator(matplotlib.dates.MonthLocator(interval=3))
        else:
            ax.xaxis.set_major_locator(matplotlib.dates.MonthLocator())
        ax.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%b-%Y'))
        # Rotate
        plt.setp( ax.xaxis.get_majorticklabels(), rotation=45 )
        plt.setp( ax.xaxis.get_majorticklabels(), fontsize='x-small' )
    
    if plottype == PLOT_TYPE_CDF:
        ax.set_xlim(min_x, max_x)
        ax.set_xscale('log')
        if args.xlabel:
            ax.set_xlabel(args.xlabel)
        else:
            pass
            ax.set_xlabel( 'Streamflow ($mm\ day^{-1}$)' )
    elif args.xlabel:
        ax.set_xlabel(args.xlabel)
    
    # Y-axis
    if plottype == PLOT_TYPE_LOGY:
        ax.set_yscale('log')
    
    if args.ylabel:
        ax.set_ylabel(args.ylabel)
    elif plottype != PLOT_TYPE_CDF and plottype != PLOT_TYPE_LOGY:
        y_label = 'Streamflow ($mm\ day^{-1}$)'
        ax.set_ylabel( y_label )
    
    data_plt.insert(0, obs_plt)
    
    # Plot secondary data (if specified)
    if secondary and \
        (plottype == PLOT_TYPE_STD or plottype == PLOT_TYPE_LOGY):
        # Plot
        ax2 = ax.twinx()
        if args.secondaryPlotType == 'line':
            (sec_plot,) = ax2.plot(x, data[secondary])
        elif args.secondaryPlotType == 'bar':
            sec_plot = ax2.bar(x, data[secondary], facecolor=second_color, edgecolor='none', width=2.0)
        ax2.invert_yaxis()
        ax2.set_ylabel('Precipication ($mm\ day^{-1}$)')
    
    return data_plt

if __name__ == "__main__":
    # Handle command line options
    parser = argparse.ArgumentParser(description='Plot CDF of N datasets vs. observed data')
    parser.add_argument('-n', '--outname', required=True, 
                        help='Base name of file to output figure to.  Only specify base name of file, not extension (PDF and PNG files will be produced)')
    parser.add_argument('-o', '--obs', required=True,
                        help='File containing observed data')
    parser.add_argument('-d', '--data', required=True, #nargs='+',
                        help='One or more data files')
    parser.add_argument('-t', '--title', required=False,
                        help='Title of figure')
    parser.add_argument('-x', '--xlabel', required=False,
                        help='X-axis label')
    parser.add_argument('-y', '--ylabel', required=False,
                        help='Y-axis label')
    parser.add_argument('--figureX', required=False, type=int, default=8,
                        help='The width of the plot, in inches')
    parser.add_argument('--figureY', required=False, type=int, default=6,
                        help='The height of the plot, in inches')
    parser.add_argument('--supressObs', required=False, action='store_true',
                        help='Do not plot observed data.  Observed data will still be used for aligning timeseries')
    parser.add_argument("--color", action="store_true", required=False, default=False,
                        help="Plot in color")
    parser.add_argument('--secondaryPlotType', required=False, choices=['bar', 'line'], default='bar',
                        help='Type of plot to use for secondary data.')
    parser.add_argument('--secondaryLabel', required=False,
                        help='Label to use for seconary Y-axis')
    args = parser.parse_args()

    # Open observed data
    obs_file = open(args.obs, 'r')
    (obs_datetime, obs_data) = RHESSysOutput.readObservedDataFromFile(obs_file,
                                                                      readHour=False)
    obs_file.close()
    obs = pd.DataFrame(obs_data, index=obs_datetime, columns=['observed'])

    # Open data and align to observed
    cols = ['streamflow', 'evap', 'trans', 'precip']
    obs_align = None
    max_x = min_x = 0
    mod_file = open(args.data, 'r')
    mod_df = RHESSysOutput.readColumnsFromFile(mod_file, cols,
                                               readHour=False)
    
    # Align timeseries
    (mod_align, obs_align) = mod_df.align(obs, axis=0, join='inner')
    tmp_max_x = max(mod_align['streamflow'].max(), obs_align['observed'].max())
    if tmp_max_x > max_x:
        max_x = tmp_max_x
    min_x = max(min_x, mod_align['streamflow'].min())

    mod_file.close()

    fig = plt.figure(figsize=(args.figureX, args.figureY), dpi=80, tight_layout=True)
    ax_std = fig.add_subplot(221)
    ax_log = fig.add_subplot(222)
    ax_cdf = fig.add_subplot(223)
    ax_tab = fig.add_subplot(224)

    data_plt = plotGraph(args, PLOT_TYPE_STD, obs_align, mod_align, ['streamflow'], 
                         min_x, max_x, ax_std, secondary='precip', plotColor=args.color)
    fig.text(0.0, 1.0, '(a)')
    plotGraph(args, PLOT_TYPE_LOGY, obs_align, mod_align, ['streamflow'], 
              min_x, max_x, ax_log, plotColor=args.color)
    fig.text(1.0, 1.0, '(b)')
    plotGraph(args, PLOT_TYPE_CDF, obs_align, mod_align, ['streamflow'], 
              min_x, max_x, ax_cdf, plotColor=args.color)
    fig.text(0.0, 0.5, '(c)')
    
    col_names = ['Observed', 'Modeled']
    plotTable(args, col_names, obs_align, mod_align, ax_tab)
    fig.text(1.0, 0.5, '(d)')

    # Figure annotations
    if args.title:
        fig.suptitle(args.title, y=1.01)

    legend_items = ['Observed', 'Modeled']
    fig.legend( data_plt, legend_items, 'lower right', fontsize=10, ncol=2 )

    # Output plot
    if args.color:
        outname = "%s_color" % (args.outname,)
    else:
        outname=args.outname
    plot_filename_png = "%s.png" % (outname,)
    plot_filename_pdf = "%s.pdf" % (outname,)
    plt.savefig(plot_filename_png, bbox_inches='tight') #, bbox_extra_artists=[table])
    plt.savefig(plot_filename_pdf, bbox_inches='tight') #, bbox_extra_artists=[table])



#!/usr/bin/env python
"""@package PatchZonalStatsNormalize

@brief Tool for calculating zonal statistics for normalized patch-scale RHESSys output variables.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2014, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GRASS', 'GISBASE'

2. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   grass_dbase
   grass_location
   grass_mapset


Post conditions
---------------
None

"""
import os, sys, shutil, tempfile, re
import subprocess, shlex
import time, random
import argparse
import operator
import math

import numpy as np
import statsmodels.api as sm
import matplotlib
import matplotlib.pyplot as plt

from ecohydrolib.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from ecohydrolib.grasslib import *

from rhessysworkflows.rhessys import RHESSysOutput

LINE_TYPES = ['solid', 'dashed', 'dashdot', 'dotted']
NUM_LINE_TYPES = len(LINE_TYPES)

def line_legend(ax, loc='lower right', fontsize=6, fontweight='normal',
                frameon=False):
    """ Use proxy objects to force lines in legend display
    """
    (handles, labels) = ax.get_legend_handles_labels()
    
    # Make proxy objects
    proxies = []
    for handle in handles:
        line = matplotlib.lines.Line2D([0], [0], 
                                       color=handle.get_edgecolor(),
                                       linestyle=handle.get_linestyle(),
                                       linewidth=handle.get_linewidth())
        proxies.append(line)
        
    leg = ax.legend(proxies, labels,
                    loc=loc,
                    fontsize=fontsize,
                    frameon=frameon)
    plt.setp(leg.get_texts(), fontweight=fontweight)

def simple_axis(ax, noy=False):
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    if noy:
        ax.spines['left'].set_visible(False)
        plt.setp(ax.get_yticklabels(), visible=False)
        ax.yaxis.set_major_locator(plt.NullLocator())
    else:
        ax.get_yaxis().tick_left()
    ax.get_xaxis().tick_bottom()
    

def plot_cdf(ax, data, legend_items, legend_loc='lower right', 
             numbins=1000, xlimit=None, xlabel=None, ylabel=None, title=None, linetype=None, linewidth=None, range=None,
             fontweight='normal', legend_fontsize=6, axes_fontsize=8, ticklabel_fontsize=6,
             fig_num=1, log=False):
    
    for (i, datum) in enumerate(data):
        if linetype == None:
            linestyle = LINE_TYPES[i % NUM_LINE_TYPES]
        else:
            linestyle = linetype[i]
        label = None
        if legend_items:
            label = legend_items[i]

        if log:
            log_bins = np.logspace(-1, math.log10(range[1]), numbins)
            (n, bins, patches) = \
                ax.hist(datum, bins=log_bins, label=label, normed=True, cumulative=True, stacked=False,
                        histtype='step', linestyle=linestyle, linewidth=linewidth)
            ax.set_xscale('log')
        else:
            (n, bins, patches) = \
                ax.hist(datum, numbins, range=range, label=label, normed=True, cumulative=True, stacked=False,
                        histtype='step', linestyle=linestyle, linewidth=linewidth)
        # Remove last point in graph to that the end of the graph doesn't
        # go to y=0
        patches[0].set_xy(patches[0].get_xy()[:-1])
        ax.set_ylim(0, 1)
            
    # Get rid of top and right border on axis
    simple_axis(ax)
    
    # Use lines in legend instead of boxes
    if fig_num == 1 and legend_items:
        line_legend(ax, loc=legend_loc, 
                    fontsize=legend_fontsize,
                    fontweight=fontweight)
    
    if xlabel:
        ax.set_xlabel(xlabel, fontsize=axes_fontsize, fontweight=fontweight)
    if ylabel:
        ax.set_ylabel(ylabel, fontsize=axes_fontsize, fontweight=fontweight)
    if title:
        ax.set_title(title)
    
    if xlimit:
        ax.set_xlim(xlimit[0], xlimit[1])
        ax.locator_params(axis='x', nbins=5)
    else:
        ax.set_xlim(right=bins[-2])
    plt.setp(ax.get_xticklabels(), fontsize=ticklabel_fontsize)
    plt.setp(ax.get_yticklabels(), fontsize=ticklabel_fontsize)

INT_RESCALE = 100
VARIABLE_EXPR_RE = re.compile(r'\b([a-zA-z][a-zA-Z0-9_\.]+)\b') # variable names can have "." in them
RANDOM = random.randint(100000, 999999)
RECLASS_MAP_TMP = "patchzonalstats_cover_{0}".format(RANDOM)
STATS_MAP_TMP = "patchzonalstats_output_{0}".format(RANDOM)

methods = ['average', 'mode', 'median', 'avedev', 'stddev', 'variance',
           'skewness', 'kurtosis', 'min', 'max', 'sum']

# Handle command line options
parser = argparse.ArgumentParser(description='Generate cummulative map of patch-scale RHESSys output variables')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file.')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-d', '--rhessysOutFile', required=True, nargs='+',
                    help='Directory containing RHESSys patch output, ' +\
                         'specificically patch yearly output in a file ending in "_patch.yearly".')
parser.add_argument('--mask', required=False, default=None,
                    help='Name of raster to use as a mask')
parser.add_argument('-z', '--zones', required=True, nargs='+',
                    help='Name of raster to use as zones in which statistic is to be calculated')
parser.add_argument('-o', '--outputDir', required=True,
                    help='Directory to which map should be output')
parser.add_argument('-f', '--outputFile', required=True,
                    help='Name of file to store plot in; ".pdf" will be appended. If file exists it will be overwritten.')
parser.add_argument('--xlimit', required=False, nargs=2, type=float,
                    help='Specify range of data for use with X-axis of CDF plots')
parser.add_argument('--outputFileNames', required=False, nargs='+',
                    help='Names to use for each outputfile.  Will be appended to outputFile.  If not specified, legend items will be used.')
parser.add_argument('-l', '--legend', required=False, nargs='+',
                    help='Legend item labels')
parser.add_argument('--legendloc', required=False, default='lower right',
                    help='Valid Matplotlib legend location (e.g. "lower right")')
parser.add_argument('--linetype', required=False, nargs='+',
                    help='Valid Matplotlib line type, e.g. solid, dashed, dashdot, dotted.  Must specify 1 or N, where N==number of data files')
parser.add_argument('--linewidth', required=False, type=float)
parser.add_argument('--fontweight', required=False, 
                        choices=['ultralight','light','normal','regular','book','medium','roman','semibold','demibold','demi','bold','heavy','extra bold','black'],
                        default='regular')
parser.add_argument('--legendfontsize', required=False, type=float,
                    default=6)
parser.add_argument('--axesfontsize', required=False, type=float,
                    default=8)
parser.add_argument('--ticklabelfontsize', required=False, type=float,
                    default=6)
parser.add_argument('--patchMap', required=False, default='patch',
                    help='Name of patch map')
parser.add_argument('-y', '--year', required=False, type=int,
                    help='Year for which statistics should be generated')
parser.add_argument('-v', '--outputVariable', required=True,
                    help='Name of RHESSys variable to be mapped.  Can be an expression such as "trans_sat + trans_unsat"')
parser.add_argument('-c', '--constant', required=False, nargs='+', type=float,
                    help='Constant to add to variable')
parser.add_argument('-n' ,'--variableName', required=False, nargs='+',
                    help='Names to use for variables.  If not supplied, outputVariable will be used.' +
                    'Note, do not use a dash "-" in this name.')
parser.add_argument('--normalizeMap', required=True,
                    help='Map to normalize values by (variable will be divided by this map)')
parser.add_argument('-s', '--statistic', required=True,
                    choices=methods,
                    help='Statistic to calculate')
parser.add_argument('--keepmap', required=False, action='store_true', default=True,
                    help='Whether to keep resulting zonal statistics GRASS map')
parser.add_argument('--mapcolorstyle', required=False, default='grey1.0',
                    help='Color map style to pass to r.colors, used for zonal stats map.')
parser.add_argument('--log', required=False, action='store_true',
                    help='Plot CDF/histogram with log-scaled bins')
args = parser.parse_args()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile)

if not args.legend and not args.outputFileNames:
    sys.exit('Must specify one of --legend or --outputFileNames')

if args.legend and len(args.rhessysOutFile) != len(args.legend):
    sys.exit('Number of legend items must equal the number of data files')
if args.outputFileNames and len(args.outputFileNames) != len(args.rhessysOutFile):
    sys.exit('Number of output file names must equal the number of data files')
if args.variableName and len(args.zones) != len(args.variableName):
    sys.exit('Number of variable name must equal the number of zones') 
if args.constant and len(args.constant) != len(args.rhessysOutFile):
    sys.exit('Number of constants must equal the number of data files')
if args.xlimit:
    if args.xlimit[0] >= args.xlimit[1]:
        sys.exit('First value of xlimit must be smaller than the second')

linestyles = None
if args.linetype:
    if len(args.linetype) == 1:
        linestyles = [args.linetype] * len(args.rhessysOutFile)
    elif len(args.linetype) != len(args.rhessysOutFile):
        linestyles = args.linetype

outputFileNames = args.legend
if args.outputFileNames:
    outputFileNames = args.outputFileNames

# Check for necessary information in metadata
metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))

patchFilepaths = []
for outfile in args.rhessysOutFile:
    if not os.path.isfile(outfile) or not os.access(outfile, os.R_OK):
        sys.exit("Unable to read RHESSys output file %s" % (outfile,))
    patchFilepaths.append( os.path.abspath(outfile) ) 

if not os.path.isdir(args.outputDir) or not os.access(args.outputDir, os.W_OK):
    sys.exit("Unable to write to output directory %s" % (args.outputDir,) )
outputDir = os.path.abspath(args.outputDir)
outputFile = "{0}.pdf".format(args.outputFile)
outputFilePath = os.path.join(outputDir, outputFile)

if args.variableName:
    variableLabels = args.variableName
else:
    variableLabels = [args.outputVariable] * len(args.zones)

# Determine output variables
variables = ['patchID']
m = VARIABLE_EXPR_RE.findall(args.outputVariable)
if m:
    variables += m
else:
    sys.exit("No output variables specified")

# 1. Get tmp folder for temprarily storing rules
tmpDir = tempfile.mkdtemp()
print("Temp dir: %s" % (tmpDir,) )
reclassRule = os.path.join(tmpDir, 'reclass.rule')

# 2. Initialize GRASS
grassDbase = os.path.join(context.projectDir, metadata['grass_dbase'])
grassConfig = GRASSConfig(context, grassDbase, metadata['grass_location'], metadata['grass_mapset'])
grassLib = GRASSLib(grassConfig=grassConfig)

# Set mask (if present)
if args.mask:
    result = grassLib.script.run_command('r.mask',
                                         input=args.mask,
                                         maskcats="1",
                                         flags="o")
    if result != 0:
        sys.exit("Failed to set mask using layer %s" % (args.mask,) )
    # Zoom to map extent
    result = grassLib.script.run_command('g.region',
                                         rast='MASK',
                                         zoom='MASK')
    if result != 0:
        sys.exit("Failed to set region to layer %s" % \
                 (args.mask,) )

# 3. For each rhessys output file ...
variablesList = []
for (i, patchFilepath) in enumerate(patchFilepaths):
    print("\nReading RHESSys output %s from %s  (this may take a while)...\n" \
          % (os.path.basename(patchFilepath), os.path.dirname(patchFilepath)) )
    data = np.genfromtxt(patchFilepath, names=True)
    if len(data) < 1:
        sys.exit("No data found for variable in RHESSys output file '%s'" % \
                 (patchFilepath,) )
    
    if args.year:
        data = data[data['year'] == float(args.year)]
    
    patchIDs = [ int(p) for p in data['patchID'] ]
    expr = VARIABLE_EXPR_RE.sub(r'data["\1"]', args.outputVariable)
    var = eval(expr)
    if args.constant:
        var += args.constant[i]
    variablesList.append(var)
        
# Write normalized maps for each input file, for each zone
zones = args.zones
data = {}
maps_to_delete = set()
maps_to_delete.add(STATS_MAP_TMP)
for zone in zones:
    for (i, variable) in enumerate(variablesList):
        reclass_map = "{0}_{1}".format(RECLASS_MAP_TMP, outputFileNames[i])
        # 4. Write reclass rule to temp file
        reclass = open(reclassRule, 'w') 
        for (j, var) in enumerate(variable):
            reclass.write("%d:%d:%d:%d\n" % (patchIDs[j], patchIDs[j], var, var) )
        reclass.close()
            
        # 5. Generate map for variable (only do it once per variable)
        if not reclass_map in maps_to_delete:
            print("\nMapping variable: {0} for intput {1} ...".format(args.outputVariable, outputFileNames[i]))
            result = grassLib.script.run_command('r.recode', 
                                                 input=args.patchMap, 
                                                 output=reclass_map,
                                                 rules=reclassRule,
                                                 overwrite=True)
            if result != 0:
                sys.exit("Failed to create reclass map for output: {0}".format(outputFilePath) )
            
            print("Normalizing {0} map by {1}".format(reclass_map, args.normalizeMap))
            # Normalize (Rescale to integer in the process)
            rMapcalcExpr = '$reclass=int(($reclass/$norm)*$scale)'
            grassLib.script.raster.mapcalc(rMapcalcExpr, reclass=reclass_map, norm=args.normalizeMap,
                                           scale=INT_RESCALE, verbose=True)
            
            maps_to_delete.add(reclass_map)
        
        # 6. Calculate zonal statistics
        print("Calculating zonal statistics...")
        result = grassLib.script.run_command('r.statistics',
                                             base=zone,
                                             cover=reclass_map,
                                             method=args.statistic,
                                             output=STATS_MAP_TMP,
                                             overwrite=True)
        if result != 0:
            sys.exit("Failed to create zonal statisitcs for output: {0}".format(outputFilePath) )
        
        # Keep map (if applicable), re-scaling
        if args.keepmap:
            permrast = "{0}_{1}_{2}".format(args.outputFile, zone, outputFileNames[i])
            print("Saving zonal stats to permanent map {0}".format(permrast))
            rMapcalcExpr = '$permrast=@$tmprast/float($scale)'
            grassLib.script.raster.mapcalc(rMapcalcExpr, permrast=permrast, tmprast=STATS_MAP_TMP,
                                           scale=INT_RESCALE, verbose=True)
            # Set color table
            if args.mapcolorstyle:
                result = grassLib.script.run_command('r.colors', 
                                                     map=permrast,
                                                     color=args.mapcolorstyle)
                if result != 0:
                    sys.exit("Failed to modify color map")
            
        # 7. Read zonal statistics
        pipe = grassLib.script.pipe_command('r.stats', flags='ln', input=STATS_MAP_TMP)
        stats_scaled = []
        for line in pipe.stdout:
            (parcel, stat) = line.strip().split()
            stats_scaled.append(float(stat))
        stats = np.array(stats_scaled)
        stats = stats / INT_RESCALE
        print("Median for {0} for zone {1} = {2}".format(outputFileNames[i], zone, np.median(stats)))
        print("Mean for {0} for zone {1} = {2}".format(outputFileNames[i], zone, stats.mean()))
        print("Standard dev. for {0} for zone {1} = {2}".format(outputFileNames[i], zone, np.std(stats)))
        try:
            tmp_data = data[zone]
        except KeyError:
            tmp_data = []
            data[zone] = tmp_data
        tmp_data.append(stats)

min_x = max_x = 0
for datum in data.values():
    min_x = min(np.min(datum), min_x)
    max_x = max(np.max(datum), max_x)

# 8. Make plots
fig_width = 4 * len(args.zones)
fig = plt.figure(figsize=(fig_width, 3), dpi=80, tight_layout=True)

num_zones = len(zones)
for (i, zone) in enumerate(zones):  
    var_label = variableLabels[i]
    fig_num = i + 1
    
    # Make CDF
    ax = fig.add_subplot(1, num_zones, fig_num)
    plot_cdf(ax, data[zone], args.legend, legend_loc=args.legendloc, xlabel=var_label,
             linetype=linestyles, 
             linewidth=args.linewidth,
             legend_fontsize=args.legendfontsize,
             fontweight=args.fontweight,
             axes_fontsize=args.axesfontsize,
             ticklabel_fontsize=args.ticklabelfontsize,
             range=(min_x, max_x), fig_num=fig_num, log=args.log,
             xlimit=args.xlimit)
    
fig.savefig(outputFilePath, bbox_inches='tight', pad_inches=0.125)

# Make

# Cleanup
shutil.rmtree(tmpDir)
for map in maps_to_delete:
    result = grassLib.script.run_command('g.remove', rast=map)
    if result != 0:
        sys.exit("Failed to remove temporary map %s" % (map,) )



#!/usr/bin/env python
"""@package CreateGRASSLocationFromDEM

@brief Import DEM into new GRASS GIS location

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GRASS', 'GISBASE'
   'GRASS', 'MODULE_PATH'

2. The following metadata entry(ies) must be present in the manifest section of the metadata associated with the project directory:
   dem

3. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory:
   dem_srs

Post conditions
---------------
1. Will write the following entry(ies) to the RHESSys section of metadata associated with the project directory:
   grass_dbase
   grass_location
   grass_mapset
   rhessys_dir
   model_description

2. GRASS location and mapset will be created in grass_dbase

3. Will write the following entry(ies) to the GRASS section of metadata associated with the project directory:
   dem_rast

Usage:
@code
CreateGRASSLocationFromDEM.py -p /path/to/project_dir -script GRASSData -l default -m "Grass location for RHESSys model of Dead Run watershed in Baltimore, MD"
@endcode

@note EcoHydroWorkflowLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import sys
import argparse
import re

from ecohydrolib.grasslib import *

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import RHESSysPaths

# Handle command line options
parser = argparse.ArgumentParser(description='Import spatial data needed to create RHESSys worldfile into the PERMANENT mapset of a new GRASS location')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-g', '--grassDbase', dest='grassDbase', required=False,
                    help='Path within project directory of the GRASS database where the new location is to be created.')
parser.add_argument('-l', '--location', dest='location', required=False,
                    help='Name of the new GRASS location where study area data are to be imported.')
parser.add_argument('-m', '--mapset', dest='mapset', required=False,
                    help='Name of the new GRASS mapset where study area data are to be imported.')
parser.add_argument('-d', '--description', dest='description', required=True,
                    help='Description for new location')
parser.add_argument('--overwrite', dest='overwrite', action='store_true', required=False,
                    help='Overwrite existing datasets in the GRASS mapset.  If not specified, program will halt if a dataset already exists.')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 
paths = RHESSysPaths(args.projectDir)

if not args.grassDbase:
    dbase = 'GRASSData'
else:
    dbase = args.grassDbase
grassDbase = os.path.join(context.projectDir, dbase)
location = None
if args.location:
    location = args.location
mapset = None
if args.mapset:
    mapset = args.mapset

# Check for necessary information in metadata
manifest = RHESSysMetadata.readManifestEntries(context)
if not 'dem' in manifest:
    sys.exit("Metadata in project directory %s does not contain a DEM reference" % (context.projectDir,)) 

studyArea = RHESSysMetadata.readStudyAreaEntries(context)
if not 'dem_srs' in studyArea:
    sys.exit("Metadata in project directory %s does not contain DEM spatial reference" % (context.projectDir,)) 
srsPattern = re.compile('EPSG:(\d+)')
result = srsPattern.search(studyArea['dem_srs'])
if not result:
    sys.exit("Spatial reference must be specified in the form 'EPSG:<number>'")
srs = int(result.group(1))

# Set up GRASS environment
grassConfig = GRASSConfig(context, grassDbase, location, mapset, newLocation=True, overwrite=args.overwrite)
grassLib = GRASSLib(grassConfig=grassConfig)

# Create the new location
grassLib.script.core.create_location(grassConfig.dbase, grassConfig.location, epsg=srs, desc=args.description)

# Import DEM into location
demFilepath = os.path.join(context.projectDir, manifest['dem'])
#print demFilepath
result = grassLib.script.run_command('r.in.gdal', flags="e", input=demFilepath, output='dem')
if result != 0:
    sys.exit("Failed to import DEM into new GRASS dataset %s/%s, results:\n%s" % \
             (grassDbase, args.location, result) )

# Set region to DEM
result = grassLib.script.run_command('g.region', rast='dem')
if result != 0:
    sys.exit("Failed to set region to DEM")

# Update metadata
RHESSysMetadata.writeGRASSEntry(context, 'dem_rast', 'dem')
RHESSysMetadata.writeRHESSysEntry(context, 'grass_dbase', dbase)
RHESSysMetadata.writeRHESSysEntry(context, 'grass_location', grassConfig.location)
RHESSysMetadata.writeRHESSysEntry(context, 'grass_mapset', grassConfig.mapset)
RHESSysMetadata.writeRHESSysEntry(context, 'rhessys_dir', paths.rhessysDir)
RHESSysMetadata.writeRHESSysEntry(context, 'model_description', args.description)

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)


#!/usr/bin/env python
"""@package GenerateBaseStationMap

@brief Generate climate base station raster map in GRASS for a list of base stations 
using Thiessen polygons.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GRASS', 'GISBASE'
   
2. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   grass_dbase
   grass_location
   grass_mapset
   rhessys_dir
   
Post conditions
---------------
1. Will write the following entry(ies) to the RHESSys section of metadata associated with the project directory:
   basestations_text

2. Will write the following entry(ies) to the GRASS section of metadata associated with the project directory:
   basestations_rast
 
Usage:
@code
GenerateBaseStationMap.py -p /path/to/project_dir -b /path/to/base/station/file
@endcode

@note GRASS v.voronoi routine used to generate Thiessen polygons has difficulties when there are
only two points.  This may be fixed in GRASS 6.4.3.  Until then, do not use two climate stations.

@note EcoHydroWorkflowLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import os, sys, shutil
import argparse

from ecohydrolib.grasslib import *

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from rhessysworkflows.rhessys import RHESSysPaths

# Handle command line options
parser = argparse.ArgumentParser(description='Generate climate base station raster map in GRASS for a list of base stations using Thiessen polygons.')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-b', '--basestationsFile', dest='basestationsFile', required=True,
                    help='Text file of the form: id|easting|northing|name')
parser.add_argument('--overwrite', dest='overwrite', action='store_true', required=False,
                    help='Overwrite existing datasets in the GRASS mapset.  If not specified, program will halt if a dataset already exists.')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if not os.path.isfile(args.basestationsFile):
    sys.exit("Specified base stations file '%s' is not a file" % (args.basestationsFile,) )
if not os.access(args.basestationsFile, os.R_OK):
    sys.exit("Specified base stations file '%s' is not readable" % (args.basestationsFile,) )

# Check for necessary information in metadata
metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))
if not 'rhessys_dir' in metadata:
    sys.exit("Metadata in project directory %s does not contain a RHESSys directory" % (context.projectDir,))

paths = RHESSysPaths(args.projectDir, metadata['rhessys_dir'])

# Get path of place to store climate data
projectDirClimDir = paths.RHESSYS_CLIM

# Set up GRASS environment
modulePath = context.config.get('GRASS', 'MODULE_PATH')
grassDbase = os.path.join(context.projectDir, metadata['grass_dbase'])
grassConfig = GRASSConfig(context, grassDbase, metadata['grass_location'], metadata['grass_mapset'])
grassLib = GRASSLib(grassConfig=grassConfig)    

# Copy base stations file into project directory
basestationsDest = os.path.join(projectDirClimDir, 'basestations.txt')
shutil.copyfile(args.basestationsFile, basestationsDest)

# Import base station points
p = grassLib.script.pipe_command('v.in.ascii', input=basestationsDest, out='basestation_points', x=2, y=3, cat=1,
                                 columns='cat int, x double precision, y double precision, name varchar(15)',
                                 overwrite=args.overwrite)
(stdoutStr, stderrStr) = p.communicate() 
result = p.returncode
if result != 0:
    sys.stdout.write(stdoutStr)
    sys.exit("\nv.in.ascii failed, returning %s" % (result,))
    
# Generate Thiessen polygons
p = grassLib.script.pipe_command('v.voronoi', input='basestation_points', output='basestation_areas', 
                                 overwrite=args.overwrite)
(stdoutStr, stderrStr) = p.communicate() 
result = p.returncode
if result != 0:
    sys.stdout.write(stdoutStr)
    sys.exit("\nv.voronoi failed, returning %s" % (result,))

# Rasterize Thiessen polygons
p = grassLib.script.pipe_command('v.to.rast', input='basestation_areas', output='basestations', 
                                 use='cat', labelcolumn='name', overwrite=args.overwrite)
(stdoutStr, stderrStr) = p.communicate() 
result = p.returncode
if result != 0:
    sys.stdout.write(stdoutStr)
    sys.exit("\nv.to.rast failed, returning %s" % (result,))

# Write metadata 
RHESSysMetadata.writeRHESSysEntry( context, 'basestations_text', paths.relpath(basestationsDest) )
RHESSysMetadata.writeGRASSEntry( context, 'basestations_rast', 'basestations' )

# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)



#!/usr/bin/env python
"""@package PatchToCumulativeValues

@brief Tool for printing cummulative values of patch-scale RHESSys output variables.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2014, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GRASS', 'GISBASE'
   'RHESSYS', 'PATH_OF_FFMPEG'

2. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   grass_dbase
   grass_location
   grass_mapset


Post conditions
---------------
None
   
@todo Add ability to read in two files, accumulate data, and write normalized map (value / max(value))   

"""
import os, sys, shutil, tempfile, re
import subprocess, shlex
import time, random
import argparse
import operator
import datetime

import numpy as np
import statsmodels.api as sm
import matplotlib.pyplot as plt

from ecohydrolib.context import Context
from rhessysworkflows.metadata import RHESSysMetadata
from ecohydrolib.grasslib import *

from rhessysworkflows.rhessys import RHESSysOutput

# Handle command line options
parser = argparse.ArgumentParser(description='Generate cummulative map of patch-scale RHESSys output variables')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file.')
parser.add_argument('-d', '--rhessysOutFile', required=True, nargs='+',
                    help='Directory containing RHESSys patch output, specificically patch daily output stored in a file whose name ends with "_patch.daily".')
parser.add_argument('-v', '--outputVariables', required=True, nargs='+',
                    help='Name of RHESSys variables to be mapped.  Should be a list of individual variable names, not expressions thereof')
parser.add_argument('-p', '--patchIDs', required=True, nargs='+', type=int,
                    help='Patches for which output should be produced')
parser.add_argument('--startdate', type=int, nargs=4,
                    help='Date on which to begin output, of format YYYY M D H')
parser.add_argument('--enddate', type=int, nargs=4,
                    help='Date on which to end output, of format YYYY M D H')
args = parser.parse_args()

startDate = None
if args.startdate:
    # Set end date based on command line
    startDate = datetime.datetime(args.startdate[0],
                                  args.startdate[1],
                                  args.startdate[2],
                                  args.startdate[3])

endDate = None
if args.enddate:
    # Set end date based on command line
    endDate = datetime.datetime(args.enddate[0],
                                args.enddate[1],
                                args.enddate[2],
                                args.enddate[3])

patchDailyFilepaths = []
for outfile in args.rhessysOutFile:
    if not os.path.isfile(outfile) or not os.access(outfile, os.R_OK):
        sys.exit("Unable to read RHESSys output file %s" % (outfile,))
    patchDailyFilepaths.append( os.path.abspath(outfile) ) 

# Determine output variables
variables = ['patchID'] + args.outputVariables

sys.stdout.write('scenario,patchid')
for var in args.outputVariables:
    sys.stdout.write(',sum_' + var)
sys.stdout.write('\n')

# For each rhessys output file ...
for (i, patchDailyFilepath) in enumerate(patchDailyFilepaths):
    
    scenario = os.path.basename( os.path.dirname(patchDailyFilepath) )
    
    f = open(patchDailyFilepath)
    data = RHESSysOutput.readColumnsFromPatchDailyFile(f, variables)
    f.close()
    if len(data) < 1:
        sys.exit("No data found for variable in RHESSys output file '%s'" % \
                 (patchDailyFilepath,) )
    
    # For each day
    variablesDict = {}
    patchIDs = None
    variable = None
    for (i, key) in enumerate(data):
        if startDate and key < startDate:
            continue
        if endDate and key > endDate:
            break
        #import pdb; pdb.set_trace()
        dataForDate = data[key]
        vars = {}
        for var in args.outputVariables:
            vars[var] = np.array( dataForDate[var] )
        patchIDs = np.array( [ int(f) for f in dataForDate['patchID'] ] )
        
        # Store value for each selected patchID
        for p in args.patchIDs:
            try:
                dataForPatch = variablesDict[p]
            except KeyError:
                dataForPatch = {}
                variablesDict[p] = dataForPatch
            patchIdx = np.where( patchIDs == p )
            
            for var in args.outputVariables:
                variable = vars[var]
                try:
                    tmp_data = dataForPatch[var]
                except KeyError:
                    tmp_data = []
                    dataForPatch[var] = tmp_data
                tmp_data.append( variable[patchIdx] )
    
    # Print summary for each patch
    for p in args.patchIDs:
        dataForPatch = variablesDict[p]
        sys.stdout.write("%s,%d" % (scenario, p) )
        for var in args.outputVariables:
            p_tmp = np.array( dataForPatch[var] )
            sys.stdout.write(",%f" % (  p_tmp.sum() ) )
        sys.stdout.write('\n')
    



#!/usr/bin/env python
"""@package DelineateWatershed

@brief Use GRASS GIS to delineate a watershed for study area defined by a DEM

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GRASS', 'GISBASE'

2. The following metadata entry(ies) must be present in the RHESSys section of the metadata associated with the project directory:
   grass_dbase
   grass_location
   grass_mapset

3. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory:
   dem_rows
   dem_srs
   gage_lat_wgs84
   gage_lon_wgs84
   
4. The following metadata entry(ies) must be present in the GRASS section of the metadata associated with the project directory:
   dem_rast
   
5. The following metadata entry(ies) will be used if present in the GRASS section of the metadata associated with the project directory:
   stream_burned_dem_rast

Post conditions
---------------
1. Will write the following entry(ies) to the RHESSys section of metadata associated with the project directory:
   gage_easting_raw (raw gage X coordinate in study area spatial reference system coordinates)
   gage_northing_raw (raw gage Y coordinate in study area spatial reference system coordinates)
   gage_easting_snapped (gage X coordinate, snapped to nearest stream pixel, in study area spatial reference system coordinates)
   gage_northing_snapped (gage Y coordinate, snapped to nearest stream pixel, in study area spatial reference system coordinates)
   watershed_threshold
   watershed_area_km
   watershed_area_estimate_km
    

2. Will write the following entry(ies) to the GRASS section of metadata associated with the project directory:
   gage_vect
   gage_snapped_vect
   drain_rast [drainage direction map]
   uaa_rast [upslope accumulated area]
   basin_rast [delineated watershed boundary]
   subbasins_rast [sub basins within the watershed boundary]
   hillslope_rast [hillslopes within the watershed boundary]
   zone_rast [climate zones, same as hillslope_rast]
   streams_rast [streams within the watershed boundary]
   stream_reaches_rast [streams within the watershed boundary, labeled by reach ID]
   east_horizon_rast
   west_horizon_rast
   slope_rast
   aspect_rast 
   K_rast [default Ksat map]
   m_rast [default decay of Ksat with depth map]
   zero_rast [Map where all cells are set to zero]
   one_rast [Map where all cells are set to one]
   xmap_rast
   ymap_rast
   

Usage:
@code
DelineateWatershedForGRASSLocation.py -p /path/to/project_dir -t 500
@endcode

@note EcoHydroWorkflowLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 

@todo fill pits in DEM with GRASS tool, have an argument to turn it off
@todo Factor out gage snapping into a separate script?
"""
import os, sys, errno
import argparse
import ConfigParser
import re
import textwrap

from ecohydrolib.spatialdata.utils import transformCoordinates
from ecohydrolib.grasslib import *

from rhessysworkflows.context import Context
from rhessysworkflows.metadata import RHESSysMetadata

AREA_THRESHOLD = 0.2

def positive_odd_integer(string):
    msg = "%s is not an integer >=3" % (string,)
    value = int(string)
    if value < 2:
        raise argparse.ArgumentTypeError(msg)
    elif value % 2 == 0:
        raise argparse.ArgumentTypeError(msg)
    return value
        

# Handle command line options
parser = argparse.ArgumentParser(description='Delineate watershed using GRASS GIS')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file. Must define section "GRASS" and option "GISBASE"')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-t', '--threshold', dest='threshold', required=True, type=int,
                    help='Minimum size (in cells the size of the DEM resolution) of watershed sub-basins')
parser.add_argument('-s', '--streamThreshold', dest='streamThreshold', required=False, type=float,
                    help='Threshold to pass to r.findtheriver for distinguishing stream from non-stream pixels')
parser.add_argument('-w', '--streamWindow', dest='streamWindow', required=False, type=positive_odd_integer,
                    help='Stream search window, must be a postive, odd integer')
parser.add_argument('-a', '--areaEstimate', dest='areaEstimate', required=False, type=float,
                    help='Estimated area, in sq. km, of watershed to be delineated.  A warning message will be displayed if the delineated basin area is not close to estimated area.')
parser.add_argument('--ignoreBurnedDEM', dest='ignoreBurnedDEM', action='store_true', required=False,
                    help='Ignore stream burned DEM, if present. Default DEM raster will be used for all operations. If not specified and if stream burned raster is present, stream burned DEM will be used for calculating flow direction maps.')
parser.add_argument('--multiflowdirection', dest='multiflowdirection', action='store_true', required=False,
                    help='Use multiflow direction to determine watershed boundaries.')
parser.add_argument('--overwrite', dest='overwrite', action='store_true', required=False,
                    help='Overwrite existing datasets in the GRASS mapset.  If not specified, program will halt if a dataset already exists.')
args = parser.parse_args()
cmdline = RHESSysMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

# Check for necessary information in metadata
metadata = RHESSysMetadata.readRHESSysEntries(context)
if not 'grass_dbase' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS Dbase" % (context.projectDir,)) 
if not 'grass_location' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS location" % (context.projectDir,)) 
if not 'grass_mapset' in metadata:
    sys.exit("Metadata in project directory %s does not contain a GRASS mapset" % (context.projectDir,))

studyArea = RHESSysMetadata.readStudyAreaEntries(context)
if not 'dem_srs' in studyArea:
    sys.exit("Metadata in project directory %s does not contain DEM spatial reference" % (context.projectDir,)) 
if not 'gage_lat_wgs84' in studyArea:
    sys.exit("Metadata in project directory %s does not contain gage latitude" % (context.projectDir,)) 
if not 'gage_lon_wgs84' in studyArea:
    sys.exit("Metadata in project directory %s does not contain gage longitude" % (context.projectDir,)) 
    
grassMetadata = RHESSysMetadata.readGRASSEntries(context)
if not 'dem_rast' in grassMetadata:
    sys.exit("Metadata in project directory %s does not contain a DEM raster in a GRASS mapset" % (context.projectDir,)) 

demRast = grassMetadata['dem_rast']
demRows = int(studyArea['dem_rows'])

# Determine raster to use for flow direction raster
flowDirDem = demRast
if ('stream_burned_dem_rast' in grassMetadata) and (not args.ignoreBurnedDEM):
    flowDirDem = grassMetadata['stream_burned_dem_rast']
sys.stdout.write("Using raster named '%s' to calculate flow direction map\n" % (flowDirDem,) )

# Set up GRASS environment
modulePath = context.config.get('GRASS', 'MODULE_PATH')
grassDbase = os.path.join(context.projectDir, metadata['grass_dbase'])
grassConfig = GRASSConfig(context, grassDbase, metadata['grass_location'], metadata['grass_mapset'])
grassLib = GRASSLib(grassConfig=grassConfig)

# Make sure mask and region are properly set
result = grassLib.script.run_command('r.mask', flags='r')
if result != 0:
    sys.exit("r.mask filed, returning %s" % (result,) )
result = grassLib.script.run_command('g.region', rast=demRast)
if result != 0:
    sys.exit("g.region failed to set region to DEM, returning %s" % (result,))

# Generate drainage direction map
flags = None
if args.multiflowdirection:
    flags = 'f'
result = grassLib.script.run_command('r.watershed', 
                                     elevation=flowDirDem, drainage='drain', accumulation='uaa',
                                     overwrite=args.overwrite,
                                     flags=flags)
if result != 0:
    sys.exit("r.watershed failed creating drainage direction and uaa maps, returning %s" % (result,))

RHESSysMetadata.writeGRASSEntry(context, 'drain_rast', 'drain')
RHESSysMetadata.writeGRASSEntry(context, 'uaa_rast', 'uaa')

# Convert gage coordinates into dem_srs coordinates
(easting, northing) = transformCoordinates(studyArea['gage_lon_wgs84'],
                                           studyArea['gage_lat_wgs84'],
                                           t_srs=studyArea['dem_srs'])

RHESSysMetadata.writeRHESSysEntry(context, 'gage_easting_raw', easting)
RHESSysMetadata.writeRHESSysEntry(context, 'gage_northing_raw', northing)

point = "%f|%f\n" % (easting, northing) 
result = grassLib.script.write_command('v.in.ascii', output='gage', stdin=point, overwrite=args.overwrite)
if result != 0:
    sys.exit("v.in.ascii failed to create 'gage' vector, returning %s" % (result,))

RHESSysMetadata.writeGRASSEntry(context, 'gage_vect', 'gage')

# Snap the gage to the stream
findTheRiver = os.path.join(modulePath, 'r.findtheriver')
if args.streamWindow:
    result = grassLib.script.read_command(findTheRiver, verbose=True,
                                          accumulation="uaa", easting=easting, northing=northing,
                                          threshold=args.streamThreshold,
                                          window=args.streamWindow)
else:
    result = grassLib.script.read_command(findTheRiver, verbose=True,
                                          accumulation="uaa", easting=easting, northing=northing,
                                          threshold=args.streamThreshold)
if None == result or '' == result:
    sys.stdout.write("r.findtheriver did not find a better stream pixel, using raw gage coordinate\n")

snappedCoords = result.split();
if len(snappedCoords) == 2:
    # We found the stream, update coordinates
    easting = float(snappedCoords[0])
    northing = float(snappedCoords[1])
    
RHESSysMetadata.writeRHESSysEntry(context, 'gage_easting_snapped', easting)
RHESSysMetadata.writeRHESSysEntry(context, 'gage_northing_snapped', northing)

point = "%f|%f\n" % (easting, northing) 
result = grassLib.script.write_command('v.in.ascii', output='gage_snapped', stdin=point, overwrite=args.overwrite)
if result != 0:
    sys.exit("v.in.ascii failed to create 'gage_snapped' vector, returning %s" % (result,))

RHESSysMetadata.writeGRASSEntry(context, 'gage_snapped_vect', 'gage_snapped')

# Delineate watershed
basinName = 'basin'
result = grassLib.script.run_command('r.water.outlet', drainage='drain', basin=basinName, 
                                     easting=easting, northing=northing, overwrite=args.overwrite)

if result != 0:
    sys.exit("r.water.outlet failed to delineate watershed basin, returning %s" % (result,))

RHESSysMetadata.writeGRASSEntry(context, 'basin_rast', basinName)

# Generate hillslopes
#   We have to place these options in a dictionary because one of the options
#   has a '.' in its name.
flags = None
if args.multiflowdirection:
    flags = 'f'
rWatershedOptions = {'elevation': flowDirDem, 
                     'threshold': args.threshold,
                     'basin': 'subbasins',
                     'half.basin': 'hillslopes',
                     'stream': 'streams',
                     'overwrite': args.overwrite,
                     'flags': flags }
result = grassLib.script.run_command('r.watershed', **rWatershedOptions)
if result != 0:
    sys.exit("r.watershed failed creating subbasins, returning %s" % (result,))

# Make binary stream layer
# First save stream segment raster
result = grassLib.script.run_command('g.copy', rast='streams,stream_reaches')
if result != 0:
    sys.exit("Failed to save stream reach raster, g.copy returned %s" % (result,))

result = grassLib.script.write_command('r.mapcalc', stdin='streams=(streams >= 0)')
if result != 0:
    sys.exit("r.mapcalc failed to create binary streams, returning %s" % (result,))

RHESSysMetadata.writeGRASSEntry(context, 'subbasins_rast', 'subbasins')
RHESSysMetadata.writeGRASSEntry(context, 'hillslope_rast', 'hillslopes')
RHESSysMetadata.writeGRASSEntry(context, 'stream_reaches_rast', 'stream_reaches')
RHESSysMetadata.writeGRASSEntry(context, 'streams_rast', 'streams')
RHESSysMetadata.writeRHESSysEntry(context, 'watershed_threshold', args.threshold)

# Generate derived terrain products
result = grassLib.script.run_command('r.horizon', flags='d', elevin=demRast, direction=0, horizon='east')
if result != 0:
    sys.exit("r.horizon failed, returning %s" % (result,))
result = grassLib.script.run_command('r.horizon', flags="d", elevin=demRast, direction=180, horizon='west')
if result != 0:
    sys.exit("r.horizon failed, returning %s" % (result,))

result = grassLib.script.write_command('r.mapcalc', stdin='east_horizon=sin(east_0) * 1000')
if result != 0:
    sys.exit("r.mapcalc failed to create east_horizon, returning %s" % (result,))
RHESSysMetadata.writeGRASSEntry(context, 'east_horizon_rast', 'east_horizon')
    
result = grassLib.script.write_command('r.mapcalc', stdin='west_horizon=sin(west_0) * 1000')
if result != 0:
    sys.exit("r.mapcalc failed to create west_horizon, returning %s" % (result,))
RHESSysMetadata.writeGRASSEntry(context, 'west_horizon_rast', 'west_horizon')

result = grassLib.script.run_command('r.slope.aspect', el=demRast, slope='slope', aspect='aspect', overwrite=args.overwrite)
if result != 0:
    sys.exit("r.slope.aspect failed, returning %s" % (result,))
RHESSysMetadata.writeGRASSEntry(context, 'slope_rast', 'slope')
RHESSysMetadata.writeGRASSEntry(context, 'aspect_rast', 'aspect')

result = grassLib.script.run_command('r.topidx', input=demRast, out='wetness_index', overwrite=args.overwrite)
if result != 0:
    sys.exit("r.topidx failed, returning %s" % (result,))
RHESSysMetadata.writeGRASSEntry(context, 'wetness_index_rast', 'wetness_index')

# Generate default K, m, zero, one (default for roads), and patch maps
# First set mask
result = grassLib.script.run_command('r.mask', flags="o", input="basin", maskcats=1)
if result != 0:
    sys.exit("r.mask failed, returning %s" % (result,))

result = grassLib.script.write_command('r.mapcalc', stdin='K=basin*2')
if result != 0:
    sys.exit("r.mapcalc failed to create K map, returning %s" % (result,))
RHESSysMetadata.writeGRASSEntry(context, 'K_rast', 'K')
    
result = grassLib.script.write_command('r.mapcalc', stdin='m=basin*0.12')
if result != 0:
    sys.exit("r.mapcalc failed to create m map, returning %s" % (result,))
RHESSysMetadata.writeGRASSEntry(context, 'm_rast', 'm')
    
result = grassLib.script.write_command('r.mapcalc', stdin='zero=0')
if result != 0:
    sys.exit("r.mapcalc failed to create zero map, returning %s" % (result,))
RHESSysMetadata.writeGRASSEntry(context, 'zero_rast', 'zero')

result = grassLib.script.write_command('r.mapcalc', stdin='one=1')
if result != 0:
    sys.exit("r.mapcalc failed to create one map, returning %s" % (result,))
RHESSysMetadata.writeGRASSEntry(context, 'one_rast', 'one')

result = grassLib.script.write_command('r.mapcalc', stdin='ymap=y()')
if result != 0:
    sys.exit("r.mapcalc failed to create ymap raster, returning %s" % (result,))
RHESSysMetadata.writeGRASSEntry(context, 'ymap_rast', 'ymap')
    
result = grassLib.script.write_command('r.mapcalc', stdin='xmap=x()')
if result != 0:
    sys.exit("r.mapcalc failed to create xmap raster, returning %s" % (result,))
RHESSysMetadata.writeGRASSEntry(context, 'xmap_rast', 'xmap')

# Check delineated basin area
if args.areaEstimate:
    area = -1.0
    areaRegex = re.compile('^\|1\|.*\|(.*)\|$')
    # Get delineated area in hectares
    pipe = grassLib.script.pipe_command('r.report', map=basinName, units='k')
    for line in pipe.stdout:
        m = areaRegex.match(line)
        if m:
            area = float(m.group(1))
            if abs(area - args.areaEstimate) / args.areaEstimate > AREA_THRESHOLD:
                sys.stdout.write(textwrap.fill("WARNING: Delineated area of %.2f sq. km differs from estimated area %.2f sq. km by MORE than %.0f%%. Try increasing or decreasing the stream threshold used for gage snapping.\n" % \
                                   (area, args.areaEstimate, AREA_THRESHOLD * 100) ) )
            else:
                sys.stdout.write(textwrap.fill("OK: Delineated area of %.2f sq. km differs from estimated area %.2f sq. km by less than %.0f%%\n" % \
                                   (area, args.areaEstimate, AREA_THRESHOLD * 100) ) )
    # Write metadata
    RHESSysMetadata.writeRHESSysEntry(context, 'watershed_area_km', area)
    RHESSysMetadata.writeRHESSysEntry(context, 'watershed_area_estimate_km', args.areaEstimate)

sys.stdout.write('\n\nFinished delineating watershed\n')
                
# Write processing history
RHESSysMetadata.appendProcessingHistoryItem(context, cmdline)



"""@package rhessysworkflows.rhessys
    
@brief Functions related to RHESSys models

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>

@todo Add unit tests
"""
import os, errno
import string
from datetime import datetime
from datetime import timedelta
from collections import OrderedDict

import pandas as pd

from rhessysworkflows.metadata import RHESSysMetadata


class RHESSysOutput(object):
    
    TIME_STEP_HOURLY = 1
    TIME_STEP_DAILY = 2
    TIME_STEPS = [TIME_STEP_HOURLY, TIME_STEP_DAILY]
    
    HOUR_HEADER = 'hour'
    DAY_HEADER = 'day'
    MONTH_HEADER = 'month'
    YEAR_HEADER = 'year'
    
    @classmethod
    def readObservedDataFromFile(cls, f, header=True, timeStep=TIME_STEP_DAILY, logger=None,
                                 readHour=True):
        """ Reads the data from the observed data file.  Assumes that
            there is one data point per line.  By default a daily timestep
            is assumed, but hourly is also supported; time step is used for
            calculating date for each datum.
            
            Arguments:
            f -- file object     The text file to read from
            header -- boolean    Specifies whether a header is present
                                  in the file.  If True, the first
                                  line of the file will read and used to
                                  determine start date of the timeseries.
                                  Date is assumed to be in the format:
                                  "YYYY M D H"
            timeStep -- string   One of RHESSysOutput.TIME_STEPS
            readHour -- boolean  Control whether the hour should be read from the file header
        
            Returns tuple (list<datetime.datetime>, list<float>)
            Returns tuple (empty list, list<float>) if header is false  
            Returns tuple of empty lists if there were no data.
        """
        assert(timeStep in RHESSysOutput.TIME_STEPS)
        
        date_list = []
        obs_data = []
        tmpDate = None
        delta = None
        
        if header:
            headerData = f.readline().split()
            if readHour:
                tmpDate = datetime(int(headerData[0]), int(headerData[1]), 
                                   int(headerData[2]), int(headerData[3]) )
            else:
                tmpDate = datetime(int(headerData[0]), int(headerData[1]), 
                                   int(headerData[2]) )
                
            if timeStep == RHESSysOutput.TIME_STEP_HOURLY:
                delta = timedelta(hours=1)
            else:
                delta = timedelta(days=1)
            if logger:
                logger.debug("Observed timeseries begin date: %s" % (str(tmpDate),) )
           
        data = f.readline()
        while data:
            obs_data.append(float(data))
            if header:
                date_list.append(tmpDate)
                tmpDate = tmpDate + delta
            
            data = f.readline()

        return (date_list, obs_data)

    @classmethod
    def readColumnsFromFile(cls, f, column_names, sep=' ', logger=None,
                            readHour=True):
        """ Reads the specified columns from the text file.  The file
            must have a header.  Reads dates/datetime from file by searching
            for headers with names of 'hour', 'day', 'month', 'year'
        
            Arguments:
            f -- file object  The text file to read from
            column_names -- A list of column names to return
            sep -- The field separator (defaults to ' ')
            readHour -- boolean  Control whether the hour should be read from the file header

            Returns Pandas DataFrame object indexed by date
            
            Raises exception if data file does not include year, month, and day fields
        """
        cols = column_names
        if readHour:
            cols = cols + [RHESSysOutput.HOUR_HEADER]
        cols = cols + [RHESSysOutput.DAY_HEADER, 
                 RHESSysOutput.MONTH_HEADER, 
                 RHESSysOutput.YEAR_HEADER]
        df = pd.read_csv(f, sep=' ', usecols=cols)
        # Build index
        time_stamps = None
        try:
            time_stamps = df[RHESSysOutput.YEAR_HEADER].apply(str)
            df = df.drop(RHESSysOutput.YEAR_HEADER, 1)
        except KeyError:
            raise Exception('Data file lacks year column')
        try:
            time_stamps += '/' + df[RHESSysOutput.MONTH_HEADER].apply(str)
            df = df.drop(RHESSysOutput.MONTH_HEADER, 1)
        except KeyError:
            raise Exception('Data file lacks month column')
        try:
            time_stamps += '/' + df[RHESSysOutput.DAY_HEADER].apply(str)
            df = df.drop(RHESSysOutput.DAY_HEADER, 1)
        except KeyError:
            raise Exception('Data file lacks day column')
        try:
            time_stamps += ' ' + df[RHESSysOutput.HOUR_HEADER].apply(str) + '00:00'
            df = df.drop(RHESSysOutput.HOUR_HEADER, 1)
        except KeyError:
            pass
        
        dates = [pd.to_datetime(date) for date in time_stamps]
        datesDf = pd.DataFrame(dates, columns=['datetime'])
        
        df = datesDf.join(df, how='inner')
        df = df.set_index('datetime')
        return df
   

    @classmethod
    def readColumnFromFile(cls, f, column_name, sep=" ", logger=None, startHour=1):
        """ Reads the specified column from the text file.  The file
            must have a header.  Reads dates/datetime from file by searching
            for headers with names of 'hour', 'day', 'month', 'year'
        
            Arguments:
            f -- file object  The text file to read from
            column_name -- The name of the column to return
            sep -- The field separator (defaults to " ")
            startHour -- Hour to use for daily data

            Returns tuple (list<datetime.datetime>, list<float>).  
            Returns tuple of empty lists if the column had no data, or if the column was
            not found
        """
        date_list = []
        col_data = []

        # Read the header line
        header = f.readline()
        headers = string.split(header, sep)
        col_idx = -1
        hour_idx = -1
        day_idx = -1
        month_idx = -1
        year_idx = -1
        col_found = False
        # Find column_name in headers
        for (counter, col) in enumerate(headers):
            if col == column_name:
                col_idx = counter
                col_found = True
            elif col == RHESSysOutput.HOUR_HEADER:
                hour_idx = counter
            elif col == RHESSysOutput.DAY_HEADER:
                day_idx = counter
            elif col == RHESSysOutput.MONTH_HEADER:
                month_idx = counter
            elif col == RHESSysOutput.YEAR_HEADER:
                year_idx = counter
            
        # We found column_name, read the data
        if col_found:
            data = f.readline()
            while data:
                hour = day = month = year = None
                cols = string.split(data, sep)
                # Get data
                col_data.append(float(cols[col_idx]))
                # Get datetime
                if hour_idx >= 0:
                    hour = int(cols[hour_idx])
                if day_idx >= 0:
                    day = int(cols[day_idx])
                if month_idx >= 0:
                    month = int(cols[month_idx])
                if year_idx >= 0:
                    year = int(cols[year_idx])
                # Construct date object
                tmpDate = None
                if hour and day and month and year:
                    tmpDate = datetime(year, month, day, hour)
                elif day and month and year:
                    tmpDate = datetime(year, month, day, startHour)
                elif month and year:
                    tmpDate = datetime(year, month, 1)
                elif year:
                    tmpDate = datetime(year, 12, 31)
                date_list.append(tmpDate)
                
                data = f.readline()

        return (date_list, col_data)

    @classmethod
    def readColumnsFromPatchDailyFile(cls, f, column_names, sep=" "):
        """ Reads the specified columns of data from a RHESSys patch daily output 
            file.  The file must have a header.  Reads dates/datetime from file by searching
            for headers with names of 'hour', 'day', 'month', 'year'
        
            Arguments:
            f -- file object  The text file to read from
            column_names -- List of the names of the columns to return
            sep -- The field separator (defaults to " ")

            Returns collection.OrderedDict<datetime.datetime, dict<string, list<float>>, 
            where the value dict for each datetime key uses column_name as its key.  
            Returns An empty dict if data for the specified columns were not found.
        """
        returnDict = OrderedDict()

        col_idx = {}
        found = False

        # Read the header line
        header = f.readline().strip()
        if ' ' == sep:
            headers = string.split(header)
        else:
            headers = string.split(header, sep)
        hour_idx = -1
        day_idx = -1
        month_idx = -1
        year_idx = -1
        # Find column_name in headers
        for (counter, col) in enumerate(headers):
            if col in column_names:
                col_idx[col] = counter
                found = True
            elif col == RHESSysOutput.HOUR_HEADER:
                hour_idx = counter
            elif col == RHESSysOutput.DAY_HEADER:
                day_idx = counter
            elif col == RHESSysOutput.MONTH_HEADER:
                month_idx = counter
            elif col == RHESSysOutput.YEAR_HEADER:
                year_idx = counter
            
        # We found column_name, read the data
        if found:
            data = f.readline().strip()
            while data and data != '':
                hour = day = month = year = None
                if ' ' == sep:
                    cols = string.split(data)
                else:
                    cols = string.split(data, sep)
                if not len(cols): break;
                # Get datetime
                if hour_idx >= 0:
                    hour = int(cols[hour_idx])
                if day_idx >= 0:
                    day = int(cols[day_idx])
                if month_idx >= 0:
                    month = int(cols[month_idx])
                if year_idx >= 0:
                    year = int(cols[year_idx])
                # Construct date object
                tmpDate = None
                if hour and day and month and year:
                    tmpDate = datetime(year, month, day, hour)
                elif day and month and year:
                    tmpDate = datetime(year, month, day, 1)
                elif month and year:
                    tmpDate = datetime(year, month, 1)
                elif year:
                    tmpDate = datetime(year, 12, 31)
                    
                try:
                    dataForDate = returnDict[tmpDate]
                except KeyError:
                    dataForDate = {}
                    returnDict[tmpDate] = dataForDate
                 
                # Get data
                for key in col_idx:
                    try:
                        tmpData = dataForDate[key]
                    except KeyError:
                        tmpData = []
                        dataForDate[key] = tmpData
                    # TODO: intelligently handle different types    
                    tmpData.append( float(cols[ col_idx[key] ]) )
                
                data = f.readline()

        return (returnDict)

def generateCommandString(binPath, outputPrefix, startDate, endDate, tecPath,
                          worldPath, subsurfaceFlowPath=None, surfaceFlowPath=None,
                          flags="", **kwargs):
    """ Return a string representing a properly formatted RHESSys command with
        the executable and all command line options and arguments specified.
        
        @param binPath String representing path to RHESSys binary
        @param startDate datetime
        @param endDate datetime
        @param tecPath String representing tec file to be used
        @param worldPath String representing world file to be used
        @param subsurfaceFlowPath String representing subsurface flowtable to be used
        @param surfaceFlowPath String representing surface flowtable to be used
        @params flags String representing flags to include (e.g. b for -b or basin output)
        @params **kwargs Mapping type describing calibration options, with key
        representing the parameter name and value a tuple of arguments to pass to the
        calibration option
        
        @return String representing the RHESSys command line 
    """
    cmd = "%s -st %s -ed %s -t %s -w %s" % \
            (binPath,
             datetimeToString(startDate), datetimeToString(endDate),
             tecPath, worldPath)
    if subsurfaceFlowPath:
        cmd = "%s -r %s" % (cmd, subsurfaceFlowPath)
    if surfaceFlowPath and subsurfaceFlowPath:
        cmd = "%s %s" % (cmd, surfaceFlowPath)
    if outputPrefix:
        cmd = "%s -pre %s" % (cmd, outputPrefix)
    # Add flags
    flagsStr = ''
    if len(flags) > 0:
        flagsStr = "-%s" % (flags[0],)
        for c in flags[1:]:
            flagSstr = "%s -%s" % (c,)
        cmd = "%s %s" % (cmd, flagsStr,)
    # Add calibration parameters
    for opt, val in kwargs.iteritems():
        valStr = ' '.join(map(str, val))
        cmd = "%s -%s %s" % (cmd, opt, valStr)
    return cmd
    
    
def datetimeToString(dt):
    """ Return string representing the year, month, day, and hour of a datetime
        object.
        
        @param dt Python datetime object
        
        @return String of the form 'YYYY MM DD HH'
    """
    return "%d %d %d %d" % \
        (dt.year, dt.month, dt.day, dt.hour)


def readParameterFile(paramFilepath):
    """ Read a RHESSys parameter file into a dictionary with
        parameters as keys and parameter values as values
        
        @param paramFilepath String representing path of parameter file to read
        @return Dict with keys representing parameters and their values
        @raise IOError if file cannot be read
    """
    if not os.access(paramFilepath, os.R_OK):
        raise IOError(errno.EACCES, "Unable to read parameter file %s" % (paramFilepath,) )
    
    paramFile = {}
    with open(paramFilepath) as f:
        for line in f:
            tokens = line.strip().split()
            if len(tokens) >= 2 and tokens[0] != '#':
                paramFile[tokens[1].lower()] = tokens[0]
    return paramFile


class RHESSysPaths(object):

    _DB = 'db'
    _ACTIVE = 'active'
    _DIR = 'rhessys'
    _SRC = 'src'
    _TEMPLATES = 'templates'
    _BIN = 'bin'
    _WORLD = 'worldfiles'
    _FLOW = 'flow'
    _TEC = 'tecfiles'
    _DEF = 'defs'
    _CLIM = 'clim'
    _OUT = 'output'
    _OBS = 'obs'
    
    def relpath(self, path):
        """ Return the portion of path relative to self.basedir
        
            @param path String representing a path within self.basedir
            @return The relative path, or None if path was not
            rooted at self.basedir
        """
        relpath = os.path.relpath(path, self.basedir)
        if relpath.find('../') != -1:
            relpath = None
        return relpath
    
    
    def getReclassRulesDirectory(self):
        """ Get path to directory, within the project directory, for storing raster reclass rules in.  
            If the directory does not exist it will be created.
            
            @return String representing the path of the rules directory
        """
        projectDirRuleDir = os.path.join(self.basedir, RHESSysMetadata.RULES_DIR)
        try:
            os.mkdir(projectDirRuleDir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise e
        return projectDirRuleDir
    
    
    def __init__(self, basedir, rhessysDir=_DIR):
            """ Verify that RHESSys directory structure is present,
                creating directories as needed.
    
                @param basedir String representing the path to the directory in which RHESSys directories will be created
                @param rhessysDir String representing root directory into which RHESSys directories will be created
    
                @raise IOError is self.basedir is not writeable.  
                @raise OSError if any directory creation fails.
            """
            self.basedir = os.path.abspath(basedir)
            self.rhessysDir = rhessysDir
            self.RHESSYS_DIR = os.path.join(self.basedir, self.rhessysDir)
            
            # Check for write access in self.basedir
            if not os.access(self.basedir, os.W_OK):
                raise IOError(errno.EACCES, "The directory %s is not writable" 
                              % self.basedir)
            # Create DB directory
            try:
                self.DB_DIR = os.path.join(self.RHESSYS_DIR, self._DB)
                os.makedirs( self.DB_DIR )
            except OSError as e:
                # If the directory exists, eat the error
                if e.errno != errno.EEXIST:
                    raise e
    
            # Create rhessys/src directory
            try:
                self.RHESSYS_SRC = os.path.join(self.RHESSYS_DIR, self._SRC)
                os.makedirs( self.RHESSYS_SRC )
            except OSError as e:
                # If the directory exists, eat the error
                if e.errno != errno.EEXIST:
                    raise e
    
            # Create rhessys/templates directory
            try:
                self.RHESSYS_TEMPLATES = os.path.join(self.RHESSYS_DIR, self._TEMPLATES)
                os.makedirs( self.RHESSYS_TEMPLATES )
            except OSError as e:
                # If the directory exists, eat the error
                if e.errno != errno.EEXIST:
                    raise e
    
            # Create rhessys/bin directory
            try:
                self.RHESSYS_BIN = os.path.join(self.RHESSYS_DIR, self._BIN)
                os.makedirs( self.RHESSYS_BIN )
            except OSError as e:
                # If the directory exists, eat the error
                if e.errno != errno.EEXIST:
                    raise e
    
            # Create rhessys/worldfiles/active directory
            try:
                self.RHESSYS_WORLD = os.path.join(self.RHESSYS_DIR, self._WORLD)
                self.RHESSYS_WORLD_ACTIVE = os.path.join(self.RHESSYS_WORLD, self._ACTIVE)
                os.makedirs( self.RHESSYS_WORLD_ACTIVE )
            except OSError as e:
                # If the directory exists, eat the error
                if e.errno != errno.EEXIST:
                    raise e
    
            # Create rhessys/flow directory
            try:
                self.RHESSYS_FLOW = os.path.join(self.RHESSYS_DIR, self._FLOW)
                os.mkdir( self.RHESSYS_FLOW )
            except OSError as e:
                # If the directory exists, eat the error
                if e.errno != errno.EEXIST:
                    raise e
    
            # Create rhessys/tecfiles/active directory
            try:
                self.RHESSYS_TEC = os.path.join(self.RHESSYS_DIR, self._TEC)
                self.RHESSYS_TEC_ACTIVE  = os.path.join(self.RHESSYS_TEC, self._ACTIVE)
                os.makedirs( self.RHESSYS_TEC_ACTIVE )
            except OSError as e:
                # If the directory exists, eat the error
                if e.errno != errno.EEXIST:
                    raise e
    
            # Create rhessys/defs directory
            try:
                self.RHESSYS_DEF = os.path.join(self.RHESSYS_DIR, self._DEF)
                os.mkdir( self.RHESSYS_DEF )
            except OSError as e:
                # If the directory exists, eat the error
                if e.errno != errno.EEXIST:
                    raise e
            
            # Create rhessys/clim directory
            try:
                self.RHESSYS_CLIM = os.path.join(self.RHESSYS_DIR, self._CLIM)
                os.mkdir( self.RHESSYS_CLIM )
            except OSError as e:
                # If the directory exists, eat the error
                if e.errno != errno.EEXIST:
                    raise e
    
            # Create output directory
            try:
                self.RHESSYS_OUT = os.path.join(self.RHESSYS_DIR, self._OUT)
                os.mkdir( self.RHESSYS_OUT )
            except OSError as e:
                # If the directory exists, eat the error
                if e.errno != errno.EEXIST:
                    raise e
    
            # Create obs directory
            try:
                self.RHESSYS_OBS = os.path.join(self.RHESSYS_DIR, self._OBS)
                os.mkdir( self.RHESSYS_OBS )
            except OSError as e:
                # If the directory exists, eat the error
                if e.errno != errno.EEXIST:
                    raise e